# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Contain utility classes for "managers", i.e., classes that act allow
to access members of other classes via TAB-completable attributes
(e.g. the class underlying `calculation.inputs` to allow to do `calculation.inputs.<label>`).
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from aiida.common.links import LinkType

__all__ = ('NodeLinksManager', 'AttributeManager')


class NodeLinksManager(object):  # pylint: disable=too-few-public-methods
    """
    A manager that allows to inspect, with tab-completion, nodes linked to a given one.
    See an example of its use in `CalculationNode.inputs`.
    """

    def __init__(self, node, link_type, incoming):
        """
        Initialise the link manager.

        :param node: the reference node object
        :param link_type: the link_type to inspect
        :param incoming: if True, inspect incoming links, otherwise inspect
            outgoing links
        """
        # This import is here to avoid circular imports
        from aiida.orm import Node

        if not isinstance(node, Node):
            raise TypeError('node must be a valid AiiDA Node')
        self._node = node
        if not isinstance(link_type, LinkType):
            raise TypeError('link_type must be a valid LinkType')
        self._link_type = link_type
        self._incoming = incoming

    def _get_keys(self):
        """Return the valid link labels, used e.g. to make getattr() work"""
        if self._incoming:
            node_attributes = self._node.get_incoming(link_type=self._link_type).all_link_labels()
        else:
            node_attributes = self._node.get_outgoing(link_type=self._link_type).all_link_labels()
        return node_attributes

    def _get_node_by_link_label(self, label):
        """
        Return the linked node with a given link label

        :param label: the link label connecting the current node to the node to get
        """
        if self._incoming:
            return self._node.get_incoming(link_type=self._link_type).get_node_by_label(label)
        return self._node.get_outgoing(link_type=self._link_type).get_node_by_label(label)

    def __dir__(self):
        """
        Allow to list all valid input links
        """
        node_attributes = self._get_keys()
        return sorted(set(list(dir(type(self))) + list(node_attributes)))

    def __iter__(self):
        node_attributes = self._get_keys()
        for k in node_attributes:
            yield k

    def __getattr__(self, name):
        """
        :param name: name of the attribute to be asked to the parser results.
        """
        try:
            return self._get_node_by_link_label(label=name)
        except KeyError:
            raise AttributeError("Node '{}' does not have an input with link '{}'".format(self._node.pk, name))

    def __getitem__(self, name):
        """
        interface to get to the parser results as a dictionary.

        :param name: name of the attribute to be asked to the parser results.
        """
        try:
            return self._get_node_by_link_label(label=name)
        except KeyError:
            raise KeyError("Node '{}' does not have an input with link '{}'".format(self._node.pk, name))

    def __str__(self):
        """Return a string representation of the manager"""
        return 'Manager for {} {} links for node pk={}'.format(
            'incoming' if self._incoming else 'outgoing', self._link_type.value.upper(), self._node.pk
        )

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, str(self))


class AttributeManager(object):  # pylint: disable=too-few-public-methods
    """
    An object used internally to return the attributes as a dictionary.
    This is currently used in :py:class:`~aiida.orm.nodes.data.dict.Dict`,
    for instance.

    :note: Important! It cannot be used to change variables, just to read
      them. To change values (of unstored nodes), use the proper Node methods.
    """

    def __init__(self, node):
        """
        :param node: the node object.
        """
        # Possibly add checks here
        self._node = node

    def __dir__(self):
        """
        Allow to list the keys of the dictionary
        """
        return sorted(self._node.attributes_items())

    def __iter__(self):
        """
        Return the keys as an iterator
        """
        for k in self._node.attributes_keys():
            yield k

    def _get_dict(self):
        """
        Return the internal dictionary
        """
        return dict(self._node.attributes_items())

    def __getattr__(self, name):
        """
        Interface to get to dictionary values, using the key as an attribute.

        :note: it works only for attributes that only contain letters, numbers
          and underscores, and do not start with a number.

        :param name: name of the key whose value is required.
        """
        return self._node.get_attribute(name)

    def __getitem__(self, name):
        """
        Interface to get to dictionary values as a dictionary.

        :param name: name of the key whose value is required.
        """
        try:
            return self._node.get_attribute(name)
        except AttributeError as err:
            raise KeyError(str(err))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,too-few-public-methods
"""Data migration for `TrajectoryData` nodes where symbol lists are moved from repository array to attribute.

This process has to be done in two separate consecutive migrations to prevent data loss in between.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module,import-error
from django.db import migrations

from aiida.backends.djsite.db.migrations import upgrade_schema_version
from aiida.backends.general.migrations import utils
from . import ModelModifierV0025

REVISION = '1.0.27'
DOWN_REVISION = '1.0.26'


def delete_trajectory_symbols_array(apps, _):
    """Delete the symbols array from all `TrajectoryData` nodes."""
    DbNode = apps.get_model('db', 'DbNode')
    DbAttribute = apps.get_model('db', 'DbAttribute')

    modifier = ModelModifierV0025(apps, DbAttribute)

    nodes = DbNode.objects.filter(type='node.data.array.trajectory.TrajectoryData.').values_list('id', 'uuid')
    for pk, uuid in nodes:
        modifier.del_value_for_node(DbNode.objects.get(pk=pk), 'array|symbols')
        utils.delete_numpy_array_from_repository(uuid, 'symbols')


def create_trajectory_symbols_array(apps, _):
    """Create the symbols array for all `TrajectoryData` nodes."""
    import numpy

    DbNode = apps.get_model('db', 'DbNode')
    DbAttribute = apps.get_model('db', 'DbAttribute')

    modifier = ModelModifierV0025(apps, DbAttribute)

    nodes = DbNode.objects.filter(type='node.data.array.trajectory.TrajectoryData.').values_list('id', 'uuid')
    for pk, uuid in nodes:
        symbols = numpy.array(modifier.get_value_for_node(pk, 'symbols'))
        utils.store_numpy_array_in_repository(uuid, 'symbols', symbols)
        modifier.set_value_for_node(DbNode.objects.get(pk=pk), 'array|symbols', list(symbols.shape))


class Migration(migrations.Migration):
    """Deleting duplicated information stored in TrajectoryData symbols numpy arrays"""

    dependencies = [
        ('db', '0026_trajectory_symbols_to_attribute'),
    ]

    operations = [
        migrations.RunPython(delete_trajectory_symbols_array, reverse_code=create_trajectory_symbols_array),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to be used as a base for data containers that represent base python data types."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import abc
import six

try:
    from functools import singledispatch  # Python 3.4+
except ImportError:
    from singledispatch import singledispatch

from .data import Data

__all__ = ('BaseType', 'to_aiida_type')


@singledispatch
def to_aiida_type(value):
    """
    Turns basic Python types (str, int, float, bool) into the corresponding AiiDA types.
    """
    raise TypeError('Cannot convert value of type {} to AiiDA type.'.format(type(value)))


@six.add_metaclass(abc.ABCMeta)
class BaseType(Data):
    """`Data` sub class to be used as a base for data containers that represent base python data types."""

    def __init__(self, *args, **kwargs):
        try:
            getattr(self, '_type')
        except AttributeError:
            raise RuntimeError('Derived class must define the `_type` class member')

        super(BaseType, self).__init__(**kwargs)

        try:
            value = args[0]
        except IndexError:
            value = self._type()  # pylint: disable=no-member

        self.value = value

    @property
    def value(self):
        return self.get_attribute('value', None)

    @value.setter
    def value(self, value):
        self.set_attribute('value', self._type(value))  # pylint: disable=no-member

    def __str__(self):
        return super(BaseType, self).__str__() + ' value: {}'.format(self.value)

    def __eq__(self, other):
        if isinstance(other, BaseType):
            return self.value == other.value
        return self.value == other

    def __ne__(self, other):
        if isinstance(other, BaseType):
            return self.value != other.value
        return self.value != other

    def new(self, value=None):
        return self.__class__(value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test utility to import, inspect, or migrate AiiDA export archives."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
import tarfile
import zipfile

from aiida.common import json
from aiida.common.exceptions import NotExistent
from aiida.tools.importexport.common.archive import extract_tar, extract_zip
from aiida.common.folders import SandboxFolder


def get_archive_file(archive, filepath=None, external_module=None):
    """
    Return the absolute path of the archive file used for testing purposes. The expected path for these files:

        aiida.backends.tests.fixtures.filepath

    :param archive: the relative filename of the archive
    :param filepath: str of directories of where to find archive (starting "/"s are irrelevant)
    :param external_module: string with name of external module, where archive can be found
    :return: absolute filepath of the archive test file
    """
    import importlib

    # Initialize
    dirpath_archive = archive

    # If the complete path has already been given, return archive immediately
    if os.path.isabs(dirpath_archive):
        return dirpath_archive

    # Add possible filepath
    if filepath and not isinstance(filepath, str):
        raise TypeError('filepath must be a string')
    elif filepath:
        # NOTE: This part should be changed, ending with:
        # dirpath_archive = os.path.join(*filepath, dirpath_archive)
        # When support for python 2 is dropped.
        # This will allow `filepath` to be a str or list/tuple,
        # and there will be no need to "take care" of misplaced dashes (`/`), ideally
        while filepath.startswith('/'):
            filepath = filepath[1:]
        dirpath_archive = os.path.join(filepath, dirpath_archive)

    # Use possible external module (otherwise use default, see above)
    if external_module and not isinstance(external_module, str):
        raise TypeError('external_module must be a string')
    elif external_module:
        # Use external module (will prepend the absolute path to `external_module`)
        external_path = os.path.dirname(os.path.realpath(importlib.import_module(external_module).__file__))

        dirpath_archive = os.path.join(external_path, dirpath_archive)
    else:
        # Add absolute path to local repo's fixtures
        dirpath_current = os.path.dirname(os.path.realpath(__file__))
        dirpath_migrate = os.path.join(dirpath_current, os.pardir, 'fixtures')

        dirpath_archive = os.path.join(dirpath_migrate, dirpath_archive)

    if not os.path.isfile(dirpath_archive):
        dirpath_parent = os.path.dirname(dirpath_archive)
        raise ValueError('archive {} does not exist in the archives directory {}'.format(archive, dirpath_parent))

    return dirpath_archive


def import_archive(archive, filepath=None, external_module=None):
    """Import a test archive that is an AiiDA export archive

    :param archive: the relative filename of the archive
    :param filepath: str of directories of where to find archive (starting "/"s are irrelevant)
    :param external_module: string with name of external module, where archive can be found
    """
    from aiida.tools.importexport import import_data

    dirpath_archive = get_archive_file(archive, filepath=filepath, external_module=external_module)

    import_data(dirpath_archive, silent=True)


def get_json_files(archive, silent=True, filepath=None, external_module=None):
    """Get metadata.json and data.json from an exported AiiDA archive

    :param archive: the relative filename of the archive
    :param silent: Whether or not the extraction should be silent
    :param filepath: str of directories of where to find archive (starting "/"s are irrelevant)
    :param external_module: string with name of external module, where archive can be found
    """
    # Get archive
    dirpath_archive = get_archive_file(archive, filepath=filepath, external_module=external_module)

    # Unpack archive
    with SandboxFolder(sandbox_in_repo=False) as folder:
        if zipfile.is_zipfile(dirpath_archive):
            extract_zip(dirpath_archive, folder, silent=silent)
        elif tarfile.is_tarfile(dirpath_archive):
            extract_tar(dirpath_archive, folder, silent=silent)
        else:
            raise ValueError('invalid file format, expected either a zip archive or gzipped tarball')

        try:
            with io.open(folder.get_abs_path('data.json'), 'r', encoding='utf8') as fhandle:
                data = json.load(fhandle)
            with io.open(folder.get_abs_path('metadata.json'), 'r', encoding='utf8') as fhandle:
                metadata = json.load(fhandle)
        except IOError:
            raise NotExistent('export archive does not contain the required file {}'.format(fhandle.filename))

    # Return metadata.json and data.json
    return metadata, data


def migrate_archive(input_file, output_file, silent=True):
    """Migrate contents using `migrate_recursively`
    This is essentially similar to `verdi export migrate`.
    However, since this command may be disabled, this function simulates it and keeps the tests working.

    :param input_file: filename with full path for archive to be migrated
    :param output_file: filename with full path for archive to be created after migration
    """
    from aiida.tools.importexport.migration import migrate_recursively

    # Unpack archive, migrate, and re-pack archive
    with SandboxFolder(sandbox_in_repo=False) as folder:
        if zipfile.is_zipfile(input_file):
            extract_zip(input_file, folder, silent=silent)
        elif tarfile.is_tarfile(input_file):
            extract_tar(input_file, folder, silent=silent)
        else:
            raise ValueError('invalid file format, expected either a zip archive or gzipped tarball')

        try:
            with io.open(folder.get_abs_path('data.json'), 'r', encoding='utf8') as fhandle:
                data = json.load(fhandle)
            with io.open(folder.get_abs_path('metadata.json'), 'r', encoding='utf8') as fhandle:
                metadata = json.load(fhandle)
        except IOError:
            raise NotExistent('export archive does not contain the required file {}'.format(fhandle.filename))

        # Migrate
        migrate_recursively(metadata, data, folder)

        # Write json files
        with io.open(folder.get_abs_path('data.json'), 'wb') as fhandle:
            json.dump(data, fhandle, indent=4)

        with io.open(folder.get_abs_path('metadata.json'), 'wb') as fhandle:
            json.dump(metadata, fhandle, indent=4)

        # Pack archive
        compression = zipfile.ZIP_DEFLATED
        with zipfile.ZipFile(output_file, mode='w', compression=compression, allowZip64=True) as archive:
            src = folder.abspath
            for dirpath, dirnames, filenames in os.walk(src):
                relpath = os.path.relpath(dirpath, src)
                for filename in dirnames + filenames:
                    real_src = os.path.join(dirpath, filename)
                    real_dest = os.path.join(relpath, filename)
                    archive.write(real_src, real_dest)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""User param type for click."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.cmdline.utils.decorators import with_dbenv


class UserParamType(click.ParamType):
    """
    The user parameter type for click.   Can get or create a user.
    """
    name = 'user'

    def __init__(self, create=False):
        """
        :param create: If the user does not exist, create a new instance (unstored).
        """
        self._create = create

    @with_dbenv()
    def convert(self, value, param, ctx):
        from aiida import orm

        results = orm.User.objects.find({'email': value})

        if not results:
            if self._create:
                return orm.User(email=value)

            self.fail("User '{}' not found".format(value), param, ctx)

        if len(results) > 1:
            self.fail("Multiple users found with email '{}': {}".format(value, results))

        return results[0]

    @with_dbenv()
    def complete(self, ctx, incomplete):  # pylint: disable=unused-argument,no-self-use
        """
        Return possible completions based on an incomplete value

        :returns: list of tuples of valid entry points (matching incomplete) and a description
        """
        from aiida import orm

        users = orm.User.objects.find()

        return [(user.email, '') for user in users if user.email.startswith(incomplete)]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module to define the (physical) constants used throughout the code."""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

# This is the precision with which AiiDA internally will store float numbers
# In particular, before storing a number (in attributes/extras),
# when going via clean_value, AiiDA will first serialize to a string with
# this precision, and then bring it back to float. This is important to
# ensure consistency between the precision of different backends, and also
# for hashing - the hashing MUST use the same precision, so that a hash computed
# after clean_value but before storing to the database will be the same as the
# hash recomputed (e.g. via `verdi rehash`) after storing and reloading the node
# from the database.
# See also discussion on GitHub issue #2631
#
# IMPORTANT! Changing this value requires to rehash the whole database!
# So if you plan to change this, think carefully, and you might need to add
# a migration.
AIIDA_FLOAT_PRECISION = 14

## PHYSICAL CONSTANTS BELOW HERE

bohr_to_ang = 0.52917720859  # pylint: disable=invalid-name
ang_to_m = 1.e-10  # pylint: disable=invalid-name
bohr_si = bohr_to_ang * ang_to_m  # pylint: disable=invalid-name
ry_to_ev = 13.6056917253  # pylint: disable=invalid-name
ry_si = 4.35974394 / 2. * 10**(-18)  # pylint: disable=invalid-name
hartree_to_ev = ry_to_ev * 2.  # pylint: disable=invalid-name
timeau_to_sec = 2.418884326155573e-17  # pylint: disable=invalid-name
invcm_to_THz = 0.0299792458  # pylint: disable=invalid-name

# Element table, from NIST (http://www.nist.gov/pml/data/index.cfm)
# Retrieved in October 2014 for atomic numbers 1-103, and in May 2016
# for atomic numbers 104-112, 114 and 116. In addition, element X is
# added to support unknown elements.
elements = {  # pylint: disable=invalid-name
    0: {
        'mass': 1.00000,
        'name': 'Unknown',
        'symbol': 'X'
    },
    1: {
        'mass': 1.00794,
        'name': 'Hydrogen',
        'symbol': 'H'
    },
    2: {
        'mass': 4.002602,
        'name': 'Helium',
        'symbol': 'He'
    },
    3: {
        'mass': 6.941,
        'name': 'Lithium',
        'symbol': 'Li'
    },
    4: {
        'mass': 9.012182,
        'name': 'Beryllium',
        'symbol': 'Be'
    },
    5: {
        'mass': 10.811,
        'name': 'Boron',
        'symbol': 'B'
    },
    6: {
        'mass': 12.0107,
        'name': 'Carbon',
        'symbol': 'C'
    },
    7: {
        'mass': 14.0067,
        'name': 'Nitrogen',
        'symbol': 'N'
    },
    8: {
        'mass': 15.9994,
        'name': 'Oxygen',
        'symbol': 'O'
    },
    9: {
        'mass': 18.9984032,
        'name': 'Fluorine',
        'symbol': 'F'
    },
    10: {
        'mass': 20.1797,
        'name': 'Neon',
        'symbol': 'Ne'
    },
    11: {
        'mass': 22.98977,
        'name': 'Sodium',
        'symbol': 'Na'
    },
    12: {
        'mass': 24.305,
        'name': 'Magnesium',
        'symbol': 'Mg'
    },
    13: {
        'mass': 26.981538,
        'name': 'Aluminium',
        'symbol': 'Al'
    },
    14: {
        'mass': 28.0855,
        'name': 'Silicon',
        'symbol': 'Si'
    },
    15: {
        'mass': 30.973761,
        'name': 'Phosphorus',
        'symbol': 'P'
    },
    16: {
        'mass': 32.065,
        'name': 'Sulfur',
        'symbol': 'S'
    },
    17: {
        'mass': 35.453,
        'name': 'Chlorine',
        'symbol': 'Cl'
    },
    18: {
        'mass': 39.948,
        'name': 'Argon',
        'symbol': 'Ar'
    },
    19: {
        'mass': 39.0983,
        'name': 'Potassium',
        'symbol': 'K'
    },
    20: {
        'mass': 40.078,
        'name': 'Calcium',
        'symbol': 'Ca'
    },
    21: {
        'mass': 44.955912,
        'name': 'Scandium',
        'symbol': 'Sc'
    },
    22: {
        'mass': 47.867,
        'name': 'Titanium',
        'symbol': 'Ti'
    },
    23: {
        'mass': 50.9415,
        'name': 'Vanadium',
        'symbol': 'V'
    },
    24: {
        'mass': 51.9961,
        'name': 'Chromium',
        'symbol': 'Cr'
    },
    25: {
        'mass': 54.938045,
        'name': 'Manganese',
        'symbol': 'Mn'
    },
    26: {
        'mass': 55.845,
        'name': 'Iron',
        'symbol': 'Fe'
    },
    27: {
        'mass': 58.933195,
        'name': 'Cobalt',
        'symbol': 'Co'
    },
    28: {
        'mass': 58.6934,
        'name': 'Nickel',
        'symbol': 'Ni'
    },
    29: {
        'mass': 63.546,
        'name': 'Copper',
        'symbol': 'Cu'
    },
    30: {
        'mass': 65.38,
        'name': 'Zinc',
        'symbol': 'Zn'
    },
    31: {
        'mass': 69.723,
        'name': 'Gallium',
        'symbol': 'Ga'
    },
    32: {
        'mass': 72.64,
        'name': 'Germanium',
        'symbol': 'Ge'
    },
    33: {
        'mass': 74.9216,
        'name': 'Arsenic',
        'symbol': 'As'
    },
    34: {
        'mass': 78.96,
        'name': 'Selenium',
        'symbol': 'Se'
    },
    35: {
        'mass': 79.904,
        'name': 'Bromine',
        'symbol': 'Br'
    },
    36: {
        'mass': 83.798,
        'name': 'Krypton',
        'symbol': 'Kr'
    },
    37: {
        'mass': 85.4678,
        'name': 'Rubidium',
        'symbol': 'Rb'
    },
    38: {
        'mass': 87.62,
        'name': 'Strontium',
        'symbol': 'Sr'
    },
    39: {
        'mass': 88.90585,
        'name': 'Yttrium',
        'symbol': 'Y'
    },
    40: {
        'mass': 91.224,
        'name': 'Zirconium',
        'symbol': 'Zr'
    },
    41: {
        'mass': 92.90638,
        'name': 'Niobium',
        'symbol': 'Nb'
    },
    42: {
        'mass': 95.96,
        'name': 'Molybdenum',
        'symbol': 'Mo'
    },
    43: {
        'mass': 98.0,
        'name': 'Technetium',
        'symbol': 'Tc'
    },
    44: {
        'mass': 101.07,
        'name': 'Ruthenium',
        'symbol': 'Ru'
    },
    45: {
        'mass': 102.9055,
        'name': 'Rhodium',
        'symbol': 'Rh'
    },
    46: {
        'mass': 106.42,
        'name': 'Palladium',
        'symbol': 'Pd'
    },
    47: {
        'mass': 107.8682,
        'name': 'Silver',
        'symbol': 'Ag'
    },
    48: {
        'mass': 112.411,
        'name': 'Cadmium',
        'symbol': 'Cd'
    },
    49: {
        'mass': 114.818,
        'name': 'Indium',
        'symbol': 'In'
    },
    50: {
        'mass': 118.71,
        'name': 'Tin',
        'symbol': 'Sn'
    },
    51: {
        'mass': 121.76,
        'name': 'Antimony',
        'symbol': 'Sb'
    },
    52: {
        'mass': 127.6,
        'name': 'Tellurium',
        'symbol': 'Te'
    },
    53: {
        'mass': 126.90447,
        'name': 'Iodine',
        'symbol': 'I'
    },
    54: {
        'mass': 131.293,
        'name': 'Xenon',
        'symbol': 'Xe'
    },
    55: {
        'mass': 132.9054519,
        'name': 'Caesium',
        'symbol': 'Cs'
    },
    56: {
        'mass': 137.327,
        'name': 'Barium',
        'symbol': 'Ba'
    },
    57: {
        'mass': 138.90547,
        'name': 'Lanthanum',
        'symbol': 'La'
    },
    58: {
        'mass': 140.116,
        'name': 'Cerium',
        'symbol': 'Ce'
    },
    59: {
        'mass': 140.90765,
        'name': 'Praseodymium',
        'symbol': 'Pr'
    },
    60: {
        'mass': 144.242,
        'name': 'Neodymium',
        'symbol': 'Nd'
    },
    61: {
        'mass': 145.0,
        'name': 'Promethium',
        'symbol': 'Pm'
    },
    62: {
        'mass': 150.36,
        'name': 'Samarium',
        'symbol': 'Sm'
    },
    63: {
        'mass': 151.964,
        'name': 'Europium',
        'symbol': 'Eu'
    },
    64: {
        'mass': 157.25,
        'name': 'Gadolinium',
        'symbol': 'Gd'
    },
    65: {
        'mass': 158.92535,
        'name': 'Terbium',
        'symbol': 'Tb'
    },
    66: {
        'mass': 162.5,
        'name': 'Dysprosium',
        'symbol': 'Dy'
    },
    67: {
        'mass': 164.93032,
        'name': 'Holmium',
        'symbol': 'Ho'
    },
    68: {
        'mass': 167.259,
        'name': 'Erbium',
        'symbol': 'Er'
    },
    69: {
        'mass': 168.93421,
        'name': 'Thulium',
        'symbol': 'Tm'
    },
    70: {
        'mass': 173.054,
        'name': 'Ytterbium',
        'symbol': 'Yb'
    },
    71: {
        'mass': 174.9668,
        'name': 'Lutetium',
        'symbol': 'Lu'
    },
    72: {
        'mass': 178.49,
        'name': 'Hafnium',
        'symbol': 'Hf'
    },
    73: {
        'mass': 180.94788,
        'name': 'Tantalum',
        'symbol': 'Ta'
    },
    74: {
        'mass': 183.84,
        'name': 'Tungsten',
        'symbol': 'W'
    },
    75: {
        'mass': 186.207,
        'name': 'Rhenium',
        'symbol': 'Re'
    },
    76: {
        'mass': 190.23,
        'name': 'Osmium',
        'symbol': 'Os'
    },
    77: {
        'mass': 192.217,
        'name': 'Iridium',
        'symbol': 'Ir'
    },
    78: {
        'mass': 195.084,
        'name': 'Platinum',
        'symbol': 'Pt'
    },
    79: {
        'mass': 196.966569,
        'name': 'Gold',
        'symbol': 'Au'
    },
    80: {
        'mass': 200.59,
        'name': 'Mercury',
        'symbol': 'Hg'
    },
    81: {
        'mass': 204.3833,
        'name': 'Thallium',
        'symbol': 'Tl'
    },
    82: {
        'mass': 207.2,
        'name': 'Lead',
        'symbol': 'Pb'
    },
    83: {
        'mass': 208.9804,
        'name': 'Bismuth',
        'symbol': 'Bi'
    },
    84: {
        'mass': 209.0,
        'name': 'Polonium',
        'symbol': 'Po'
    },
    85: {
        'mass': 210.0,
        'name': 'Astatine',
        'symbol': 'At'
    },
    86: {
        'mass': 222.0,
        'name': 'Radon',
        'symbol': 'Rn'
    },
    87: {
        'mass': 223.0,
        'name': 'Francium',
        'symbol': 'Fr'
    },
    88: {
        'mass': 226.0,
        'name': 'Radium',
        'symbol': 'Ra'
    },
    89: {
        'mass': 227.0,
        'name': 'Actinium',
        'symbol': 'Ac'
    },
    90: {
        'mass': 232.03806,
        'name': 'Thorium',
        'symbol': 'Th'
    },
    91: {
        'mass': 231.03588,
        'name': 'Protactinium',
        'symbol': 'Pa'
    },
    92: {
        'mass': 238.02891,
        'name': 'Uranium',
        'symbol': 'U'
    },
    93: {
        'mass': 237.0,
        'name': 'Neptunium',
        'symbol': 'Np'
    },
    94: {
        'mass': 244.0,
        'name': 'Plutonium',
        'symbol': 'Pu'
    },
    95: {
        'mass': 243.0,
        'name': 'Americium',
        'symbol': 'Am'
    },
    96: {
        'mass': 247.0,
        'name': 'Curium',
        'symbol': 'Cm'
    },
    97: {
        'mass': 247.0,
        'name': 'Berkelium',
        'symbol': 'Bk'
    },
    98: {
        'mass': 251.0,
        'name': 'Californium',
        'symbol': 'Cf'
    },
    99: {
        'mass': 252.0,
        'name': 'Einsteinium',
        'symbol': 'Es'
    },
    100: {
        'mass': 257.0,
        'name': 'Fermium',
        'symbol': 'Fm'
    },
    101: {
        'mass': 258.0,
        'name': 'Mendelevium',
        'symbol': 'Md'
    },
    102: {
        'mass': 259.0,
        'name': 'Nobelium',
        'symbol': 'No'
    },
    103: {
        'mass': 262.0,
        'name': 'Lawrencium',
        'symbol': 'Lr'
    },
    104: {
        'mass': 267.0,
        'name': 'Rutherfordium',
        'symbol': 'Rf'
    },
    105: {
        'mass': 268.0,
        'name': 'Dubnium',
        'symbol': 'Db'
    },
    106: {
        'mass': 271.0,
        'name': 'Seaborgium',
        'symbol': 'Sg'
    },
    107: {
        'mass': 272.0,
        'name': 'Bohrium',
        'symbol': 'Bh'
    },
    108: {
        'mass': 270.0,
        'name': 'Hassium',
        'symbol': 'Hs'
    },
    109: {
        'mass': 276.0,
        'name': 'Meitnerium',
        'symbol': 'Mt'
    },
    110: {
        'mass': 281.0,
        'name': 'Darmstadtium',
        'symbol': 'Ds'
    },
    111: {
        'mass': 280.0,
        'name': 'Roentgenium',
        'symbol': 'Rg'
    },
    112: {
        'mass': 285.0,
        'name': 'Copernicium',
        'symbol': 'Cn'
    },
    114: {
        'mass': 289.0,
        'name': 'Flerovium',
        'symbol': 'Fl'
    },
    116: {
        'mass': 293.0,
        'name': 'Livermorium',
        'symbol': 'Lv'
    },
}



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from pytz import UTC

from sqlalchemy import Column
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.types import Integer, String, DateTime

from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy.models.base import Base
from aiida.backends.sqlalchemy.utils import flag_modified
from aiida.common import timezone


class DbSetting(Base):
    __tablename__ = 'db_dbsetting'
    __table_args__ = (UniqueConstraint('key'),)
    id = Column(Integer, primary_key=True)

    key = Column(String(255), index=True, nullable=False)
    val = Column(JSONB, default={})

    # I also add a description field for the variables
    description = Column(String(255), default='', nullable=False)
    time = Column(DateTime(timezone=True), default=UTC, onupdate=timezone.now)

    def __str__(self):
        return "'{}'={}".format(self.key, self.getvalue())

    @classmethod
    def set_value(cls, key, value, with_transaction=True,
                  subspecifier_value=None, other_attribs={},
                  stop_if_existing=False):

        setting = sa.get_scoped_session().query(DbSetting).filter_by(key=key).first()
        if setting is not None:
            if stop_if_existing:
                return
        else:
            setting = cls()

        setting.key = key
        setting.val = value
        flag_modified(setting, 'val')
        setting.time = timezone.datetime.now(tz=UTC)
        if 'description' in other_attribs.keys():
            setting.description = other_attribs['description']
        setting.save()

    def getvalue(self):
        """
        This can be called on a given row and will get the corresponding value.
        """
        return self.val

    def get_description(self):
        """
        This can be called on a given row and will get the corresponding
        description.
        """
        return self.description

    @classmethod
    def del_value(cls, key, only_children=False, subspecifier_value=None):
        setting = sa.get_scoped_session().query(DbSetting).filter(key=key)
        setting.val = None
        setting.time = timezone.datetime.utcnow()
        flag_modified(setting, 'val')
        setting.save()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Controls the daemon
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import enum
import io
import os
import shutil
import socket
import tempfile

import six

from aiida.common.files import which
from aiida.manage.configuration import get_config

VERDI_BIN = which('verdi')
# Recent versions of virtualenv create the environment variable VIRTUAL_ENV
VIRTUALENV = os.environ.get('VIRTUAL_ENV', None)


class ControllerProtocol(enum.Enum):  # pylint: disable=too-few-public-methods
    """
    The protocol to use to for the controller of the Circus daemon
    """

    IPC = 0
    TCP = 1


def get_daemon_client(profile_name=None):
    """
    Return the daemon client for the given profile or the current profile if not specified.

    :param profile_name: the profile name, will use the current profile if None
    :return: the daemon client
    :rtype: :class:`aiida.engine.daemon.client.DaemonClient`
    :raises aiida.common.MissingConfigurationError: if the configuration file cannot be found
    :raises aiida.common.ProfileConfigurationError: if the given profile does not exist
    """
    config = get_config()

    if profile_name:
        profile = config.get_profile(profile_name)
    else:
        profile = config.current_profile

    return DaemonClient(profile)


class DaemonClient(object):  # pylint: disable=too-many-public-methods
    """
    Extension of the Profile which also provides handles to retrieve profile specific
    properties related to the daemon client
    """

    DAEMON_ERROR_NOT_RUNNING = 'daemon-error-not-running'
    DAEMON_ERROR_TIMEOUT = 'daemon-error-timeout'

    _DAEMON_NAME = 'aiida-{name}'
    _DEFAULT_LOGLEVEL = 'INFO'
    _ENDPOINT_PROTOCOL = ControllerProtocol.IPC

    def __init__(self, profile):
        """
        Construct a DaemonClient instance for a given profile

        :param profile: the profile instance :class:`aiida.manage.configuration.profile.Profile`
        """
        config = get_config()
        self._profile = profile
        self._SOCKET_DIRECTORY = None  # pylint: disable=invalid-name
        self._DAEMON_TIMEOUT = config.get_option('daemon.timeout')  # pylint: disable=invalid-name

    @property
    def profile(self):
        return self._profile

    @property
    def daemon_name(self):
        """
        Get the daemon name which is tied to the profile name
        """
        return self._DAEMON_NAME.format(name=self.profile.name)

    @property
    def cmd_string(self):
        """
        Return the command string to start the AiiDA daemon
        """
        from aiida.common.exceptions import ConfigurationError
        if VERDI_BIN is None:
            raise ConfigurationError(
                "Unable to find 'verdi' in the path. Make sure that you are working "
                "in a virtual environment, or that at least the 'verdi' executable is on the PATH"
            )
        return '{} -p {} devel run_daemon'.format(VERDI_BIN, self.profile.name)

    @property
    def loglevel(self):
        return self._DEFAULT_LOGLEVEL

    @property
    def virtualenv(self):
        return VIRTUALENV

    @property
    def circus_log_file(self):
        return self.profile.filepaths['circus']['log']

    @property
    def circus_pid_file(self):
        return self.profile.filepaths['circus']['pid']

    @property
    def circus_port_file(self):
        return self.profile.filepaths['circus']['port']

    @property
    def circus_socket_file(self):
        return self.profile.filepaths['circus']['socket']['file']

    @property
    def circus_socket_endpoints(self):
        return self.profile.filepaths['circus']['socket']

    @property
    def daemon_log_file(self):
        return self.profile.filepaths['daemon']['log']

    @property
    def daemon_pid_file(self):
        return self.profile.filepaths['daemon']['pid']

    def get_circus_port(self):
        """
        Retrieve the port for the circus controller, which should be written to the circus port file. If the
        daemon is running, the port file should exist and contain the port to which the controller is connected.
        If it cannot be read, a RuntimeError will be thrown. If the daemon is not running, an available port
        will be requested from the operating system, written to the port file and returned

        :return: the port for the circus controller
        """
        if self.is_daemon_running:
            try:
                with io.open(self.circus_port_file, 'r', encoding='utf8') as fhandle:
                    return int(fhandle.read().strip())
            except (ValueError, IOError):
                raise RuntimeError('daemon is running so port file should have been there but could not read it')
        else:
            port = self.get_available_port()
            with io.open(self.circus_port_file, 'w', encoding='utf8') as fhandle:
                fhandle.write(six.text_type(port))

            return port

    def get_circus_socket_directory(self):
        """
        Retrieve the absolute path of the directory where the circus sockets are stored if the IPC protocol is
        used and the daemon is running. If the daemon is running, the sockets file should exist and contain the
        absolute path of the directory that contains the sockets of the circus endpoints. If it cannot be read,
        a RuntimeError will be thrown. If the daemon is not running, a temporary directory will be created and
        its path will be written to the sockets file and returned.

        .. note:: A temporary folder needs to be used for the sockets because UNIX limits the filepath length to
            107 bytes. Placing the socket files in the AiiDA config folder might seem like the more logical choice
            but that folder can be placed in an arbitrarily nested directory, the socket filename will exceed the
            limit. The solution is therefore to always store them in the temporary directory of the operation system
            whose base path is typically short enough as to not exceed the limit

        :return: the absolute path of directory to write the sockets to
        """
        if self.is_daemon_running:
            try:
                return io.open(self.circus_socket_file, 'r', encoding='utf8').read().strip()
            except (ValueError, IOError):
                raise RuntimeError('daemon is running so sockets file should have been there but could not read it')
        else:

            # The SOCKET_DIRECTORY is already set, a temporary directory was already created and the same should be used
            if self._SOCKET_DIRECTORY is not None:
                return self._SOCKET_DIRECTORY

            socket_dir_path = tempfile.mkdtemp()
            with io.open(self.circus_socket_file, 'w', encoding='utf8') as fhandle:
                fhandle.write(six.text_type(socket_dir_path))

            self._SOCKET_DIRECTORY = socket_dir_path
            return socket_dir_path

    def get_daemon_pid(self):
        """
        Get the daemon pid which should be written in the daemon pid file specific to the profile

        :return: the pid of the circus daemon process or None if not found
        """
        if os.path.isfile(self.circus_pid_file):
            try:
                return int(io.open(self.circus_pid_file, 'r', encoding='utf8').read().strip())
            except (ValueError, IOError):
                return None
        else:
            return None

    @property
    def is_daemon_running(self):
        """
        Return whether the daemon is running, which is determined by seeing if the daemon pid file is present

        :return: True if daemon is running, False otherwise
        """
        return self.get_daemon_pid() is not None

    def delete_circus_socket_directory(self):
        """
        Attempt to delete the directory used to store the circus endpoint sockets. Will not raise if the
        directory does not exist
        """
        directory = self.get_circus_socket_directory()

        try:
            shutil.rmtree(directory)
        except OSError as exception:
            if exception.errno == 2:
                pass
            else:
                raise

    @classmethod
    def get_available_port(cls):
        """
        Get an available port from the operating system

        :return: a currently available port
        """
        open_socket = socket.socket()
        open_socket.bind(('', 0))
        return open_socket.getsockname()[1]

    def get_controller_endpoint(self):
        """
        Get the endpoint string for the circus controller. For the IPC protocol a profile specific
        socket will be used, whereas for the TCP protocol an available port will be found and
        saved in the profile specific port file

        :return: the endpoint string
        """
        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            endpoint = self.get_ipc_endpoint('controller')
        elif self._ENDPOINT_PROTOCOL == ControllerProtocol.TCP:
            endpoint = self.get_tcp_endpoint(self.get_circus_port())
        else:
            raise ValueError('invalid controller protocol {}'.format(self._ENDPOINT_PROTOCOL))

        return endpoint

    def get_pubsub_endpoint(self):
        """
        Get the endpoint string for the circus pubsub endpoint. For the IPC protocol a profile specific
        socket will be used, whereas for the TCP protocol any available port will be used

        :return: the endpoint string
        """
        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            endpoint = self.get_ipc_endpoint('pubsub')
        elif self._ENDPOINT_PROTOCOL == ControllerProtocol.TCP:
            endpoint = self.get_tcp_endpoint()
        else:
            raise ValueError('invalid controller protocol {}'.format(self._ENDPOINT_PROTOCOL))

        return endpoint

    def get_stats_endpoint(self):
        """
        Get the endpoint string for the circus stats endpoint. For the IPC protocol a profile specific
        socket will be used, whereas for the TCP protocol any available port will be used

        :return: the endpoint string
        """
        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            endpoint = self.get_ipc_endpoint('stats')
        elif self._ENDPOINT_PROTOCOL == ControllerProtocol.TCP:
            endpoint = self.get_tcp_endpoint()
        else:
            raise ValueError('invalid controller protocol {}'.format(self._ENDPOINT_PROTOCOL))

        return endpoint

    def get_ipc_endpoint(self, endpoint):
        """
        Get the ipc endpoint string for a circus daemon endpoint for a given socket

        :param endpoint: the circus endpoint for which to return a socket
        :return: the ipc endpoint string
        """
        filepath = self.get_circus_socket_directory()
        filename = self.circus_socket_endpoints[endpoint]
        template = 'ipc://{filepath}/{filename}'
        endpoint = template.format(filepath=filepath, filename=filename)

        return endpoint

    def get_tcp_endpoint(self, port=None):
        """
        Get the tcp endpoint string for a circus daemon endpoint. If the port is unspecified,
        the operating system will be asked for a currently available port.

        :param port: a port to use for the endpoint
        :return: the tcp endpoint string
        """
        if port is None:
            port = self.get_available_port()

        template = 'tcp://127.0.0.1:{port}'
        endpoint = template.format(port=port)

        return endpoint

    @property
    def client(self):
        """
        Return an instance of the CircusClient with the endpoint defined by the controller endpoint, which
        used the port that was written to the port file upon starting of the daemon

        N.B. This is quite slow the first time it is run due to the import of zmq.ssh
        in circus/utils.py in circus 0.15.0, which ultimately follows the import of CircusClient.

        :return: CircusClient instance
        """
        from circus.client import CircusClient
        return CircusClient(endpoint=self.get_controller_endpoint(), timeout=self._DAEMON_TIMEOUT)

    def call_client(self, command):
        """
        Call the client with a specific command. Will check whether the daemon is running first
        by checking for the pid file. When the pid is found yet the call still fails with a
        timeout, this means the daemon was actually not running and it was terminated unexpectedly
        causing the pid file to not be cleaned up properly

        :param command: command to call the circus client with
        :return: the result of the circus client call
        """
        from circus.exc import CallError

        if not self.get_daemon_pid():
            return {'status': self.DAEMON_ERROR_NOT_RUNNING}

        try:
            result = self.client.call(command)
        except CallError as exception:
            if str(exception) == 'Timed out.':
                return {'status': self.DAEMON_ERROR_TIMEOUT}
            raise exception

        return result

    def get_status(self):
        """
        Get the daemon running status

        :return: the client call response
        """
        command = {'command': 'status', 'properties': {'name': self.daemon_name}}

        return self.call_client(command)

    def get_numprocesses(self):
        """
        Get the number of running daemon processes

        :return: the client call response
        """
        command = {'command': 'numprocesses', 'properties': {'name': self.daemon_name}}

        return self.call_client(command)

    def get_worker_info(self):
        """
        Get workers statistics for this daemon

        :return: the client call response
        """
        command = {'command': 'stats', 'properties': {'name': self.daemon_name}}

        return self.call_client(command)

    def get_daemon_info(self):
        """
        Get statistics about this daemon itself

        :return: the client call response
        """
        command = {'command': 'dstats', 'properties': {}}

        return self.call_client(command)

    def increase_workers(self, number):
        """
        Increase the number of workers

        :param number: the number of workers to add
        :return: the client call response
        """
        command = {'command': 'incr', 'properties': {'name': self.daemon_name, 'nb': number}}

        return self.call_client(command)

    def decrease_workers(self, number):
        """
        Decrease the number of workers

        :param number: the number of workers to remove
        :return: the client call response
        """
        command = {'command': 'decr', 'properties': {'name': self.daemon_name, 'nb': number}}

        return self.call_client(command)

    def stop_daemon(self, wait):
        """
        Stop the daemon

        :param wait: boolean to indicate whether to wait for the result of the command
        :return: the client call response
        """
        command = {'command': 'quit', 'properties': {'waiting': wait}}

        result = self.call_client(command)

        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            self.delete_circus_socket_directory()

        return result

    def restart_daemon(self, wait):
        """
        Restart the daemon

        :param wait: boolean to indicate whether to wait for the result of the command
        :return: the client call response
        """
        command = {'command': 'restart', 'properties': {'name': self.daemon_name, 'waiting': wait}}

        return self.call_client(command)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Datastructures test
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import unittest


class TestNodeNumberJobResource(unittest.TestCase):
    """Unit tests for the NodeNumberJobResource class."""

    def test_init(self):
        """
        Test the __init__ of the NodeNumberJobResource class
        """
        from aiida.schedulers.datastructures import NodeNumberJobResource

        # No empty initialization
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource()

        # Missing required field
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(num_machines=1)
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=1)
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(tot_num_mpiprocs=1)

        # Standard info
        job_resource = NodeNumberJobResource(num_machines=2, num_mpiprocs_per_machine=8)
        self.assertEqual(job_resource.num_machines, 2)
        self.assertEqual(job_resource.num_mpiprocs_per_machine, 8)
        self.assertEqual(job_resource.get_tot_num_mpiprocs(), 16)
        # redundant but consistent information
        job_resource = NodeNumberJobResource(num_machines=2, num_mpiprocs_per_machine=8, tot_num_mpiprocs=16)
        self.assertEqual(job_resource.num_machines, 2)
        self.assertEqual(job_resource.num_mpiprocs_per_machine, 8)
        self.assertEqual(job_resource.get_tot_num_mpiprocs(), 16)
        # other equivalent ways of specifying the information
        job_resource = NodeNumberJobResource(num_mpiprocs_per_machine=8, tot_num_mpiprocs=16)
        self.assertEqual(job_resource.num_machines, 2)
        self.assertEqual(job_resource.num_mpiprocs_per_machine, 8)
        self.assertEqual(job_resource.get_tot_num_mpiprocs(), 16)
        # other equivalent ways of specifying the information
        job_resource = NodeNumberJobResource(num_machines=2, tot_num_mpiprocs=16)
        self.assertEqual(job_resource.num_machines, 2)
        self.assertEqual(job_resource.num_mpiprocs_per_machine, 8)
        self.assertEqual(job_resource.get_tot_num_mpiprocs(), 16)

        # wrong field name
        with self.assertRaises(TypeError):
            _ = NodeNumberJobResource(num_machines=2, num_mpiprocs_per_machine=8, wrong_name=16)

        # Examples of wrong informaton (e.g., number of machines or of nodes < 0
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_machines=0, num_mpiprocs_per_machine=8)
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_machines=1, num_mpiprocs_per_machine=0)
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_machines=1, tot_num_mpiprocs=0)
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=1, tot_num_mpiprocs=0)

        # Examples of inconsistent information
        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=8, num_machines=2, tot_num_mpiprocs=32)

        with self.assertRaises(ValueError):
            _ = NodeNumberJobResource(num_mpiprocs_per_machine=8, tot_num_mpiprocs=15)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test export file migration from old export versions to the newest"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os

from aiida import orm
from aiida.backends.testbase import AiidaTestCase
from aiida.backends.tests.utils.archives import get_archive_file, get_json_files, migrate_archive
from aiida.backends.tests.utils.configuration import with_temp_dir
from aiida.tools.importexport import import_data, EXPORT_VERSION as newest_version
from aiida.tools.importexport.migration import migrate_recursively, verify_metadata_version


class TestExportFileMigration(AiidaTestCase):
    """Test export file migrations"""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """Add variables (once) to be used by all tests"""
        super(TestExportFileMigration, cls).setUpClass(*args, **kwargs)

        # Known export file content used for checks
        cls.node_count = 25
        cls.struct_count = 2
        cls.known_struct_label = ''
        cls.known_cell = [[4, 0, 0], [0, 4, 0], [0, 0, 4]]
        cls.known_kinds = [
            {
                'name': 'Ba',
                'mass': 137.327,
                'weights': [1],
                'symbols': ['Ba']
            },
            {
                'name': 'Ti',
                'mass': 47.867,
                'weights': [1],
                'symbols': ['Ti']
            },
            {
                'name': 'O',
                'mass': 15.9994,
                'weights': [1],
                'symbols': ['O']
            },
        ]

        # Utility helpers
        cls.external_archive = {'filepath': 'archives', 'external_module': 'aiida-export-migration-tests'}
        cls.core_archive = {'filepath': 'export/migrate'}

    def setUp(self):
        """Reset database before each test"""
        super(TestExportFileMigration, self).setUp()
        self.reset_database()

    def test_migrate_recursively(self):
        """Test function 'migrate_recursively'"""
        import io
        import tarfile
        import zipfile

        from aiida.common.exceptions import NotExistent
        from aiida.common.folders import SandboxFolder
        from aiida.common.json import load as jsonload
        from aiida.tools.importexport.common.archive import extract_tar, extract_zip

        # Get metadata.json and data.json as dicts from v0.1 file archive
        # Cannot use 'get_json_files' for 'export_v0.1_simple.aiida',
        # because we need to pass the SandboxFolder to 'migrate_recursively'
        dirpath_archive = get_archive_file('export_v0.1_simple.aiida', **self.core_archive)

        with SandboxFolder(sandbox_in_repo=False) as folder:
            if zipfile.is_zipfile(dirpath_archive):
                extract_zip(dirpath_archive, folder, silent=True)
            elif tarfile.is_tarfile(dirpath_archive):
                extract_tar(dirpath_archive, folder, silent=True)
            else:
                raise ValueError('invalid file format, expected either a zip archive or gzipped tarball')

            try:
                with io.open(folder.get_abs_path('data.json'), 'r', encoding='utf8') as fhandle:
                    data = jsonload(fhandle)
                with io.open(folder.get_abs_path('metadata.json'), 'r', encoding='utf8') as fhandle:
                    metadata = jsonload(fhandle)
            except IOError:
                raise NotExistent('export archive does not contain the required file {}'.format(fhandle.filename))

            verify_metadata_version(metadata, version='0.1')

            # Migrate to newest version
            new_version = migrate_recursively(metadata, data, folder)
            verify_metadata_version(metadata, version=newest_version)
            self.assertEqual(new_version, newest_version)

    @with_temp_dir
    def test_no_node_export(self, temp_dir):
        """Test migration of export file that has no Nodes"""
        input_file = get_archive_file('export_v0.3_no_Nodes.aiida', **self.external_archive)
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Known entities
        computer_uuids = [self.computer.uuid]  # pylint: disable=no-member
        user_emails = [orm.User.objects.get_default().email]

        # Known export file content used for checks
        node_count = 0
        computer_count = 1 + 1  # localhost is always present
        computer_uuids.append('4f33c6fd-b624-47df-9ffb-a58f05d323af')
        user_emails.append('aiida@localhost')

        # Perform the migration
        migrate_archive(input_file, output_file)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Check known number of entities is present
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(), node_count)
        self.assertEqual(orm.QueryBuilder().append(orm.Computer).count(), computer_count)

        # Check unique identifiers
        computers = orm.QueryBuilder().append(orm.Computer, project=['uuid']).all()[0][0]
        users = orm.QueryBuilder().append(orm.User, project=['email']).all()[0][0]
        self.assertIn(computers, computer_uuids)
        self.assertIn(users, user_emails)

    def test_wrong_versions(self):
        """Test correct errors are raised if export files have wrong version numbers"""
        from aiida.tools.importexport.migration import MIGRATE_FUNCTIONS

        # Initialization
        wrong_versions = ['0.0', '0.1.0', '0.99']
        old_versions = list(MIGRATE_FUNCTIONS.keys())
        legal_versions = old_versions + [newest_version]
        wrong_version_metadatas = []
        for version in wrong_versions:
            metadata = {'export_version': version}
            wrong_version_metadatas.append(metadata)

        # Checks
        # Make sure the "wrong_versions" are wrong
        for version in wrong_versions:
            self.assertNotIn(
                version,
                legal_versions,
                msg="'{}' was not expected to be a legal version, legal version: {}".format(version, legal_versions)
            )

        # Make sure migrate_recursively throws a critical message and raises SystemExit
        for metadata in wrong_version_metadatas:
            with self.assertRaises(SystemExit) as exception:
                new_version = migrate_recursively(metadata, {}, None)

                self.assertIn(
                    'Critical: Cannot migrate from version {}'.format(metadata['export_version']),
                    exception.exception,
                    msg="Expected a critical statement for the wrong export version '{}', "
                    'instead got {}'.format(metadata['export_version'], exception.exception)
                )
                self.assertIsNone(
                    new_version,
                    msg='migrate_recursively should not return anything, '
                    "hence the 'return' should be None, but instead it is {}".format(new_version)
                )

    def test_migrate_newest_version(self):
        """
        Test critical message and SystemExit is raised, when an export file with the newest export version is migrated
        """
        # Initialization
        metadata = {'export_version': newest_version}

        # Check
        with self.assertRaises(SystemExit) as exception:
            new_version = migrate_recursively(metadata, {}, None)

            self.assertIn(
                'Critical: Your export file is already at the newest export version {}'.format(
                    metadata['export_version']
                ),
                exception.exception,
                msg="Expected a critical statement that the export version '{}' is the newest export version '{}', "
                'instead got {}'.format(metadata['export_version'], newest_version, exception.exception)
            )
            self.assertIsNone(
                new_version,
                msg='migrate_recursively should not return anything, '
                "hence the 'return' should be None, but instead it is {}".format(new_version)
            )

    @with_temp_dir
    def test_v02_to_newest(self, temp_dir):
        """Test migration of exported files from v0.2 to newest export version"""
        # Get export file with export version 0.2
        input_file = get_archive_file('export_v0.2.aiida', **self.external_archive)
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Perform the migration
        migrate_archive(input_file, output_file)
        metadata, _ = get_json_files(output_file)
        verify_metadata_version(metadata, version=newest_version)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Do the necessary checks
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(), self.node_count)

        # Verify that CalculationNodes have non-empty attribute dictionaries
        builder = orm.QueryBuilder().append(orm.CalculationNode)
        for [calculation] in builder.iterall():
            self.assertIsInstance(calculation.attributes, dict)
            self.assertNotEqual(len(calculation.attributes), 0)

        # Verify that the StructureData nodes maintained their (same) label, cell, and kinds
        builder = orm.QueryBuilder().append(orm.StructureData)
        self.assertEqual(
            builder.count(),
            self.struct_count,
            msg='There should be {} StructureData, instead {} were/was found'.format(
                self.struct_count, builder.count()
            )
        )
        for structures in builder.all():
            structure = structures[0]
            self.assertEqual(structure.label, self.known_struct_label)
            self.assertEqual(structure.cell, self.known_cell)

        builder = orm.QueryBuilder().append(orm.StructureData, project=['attributes.kinds'])
        for [kinds] in builder.iterall():
            self.assertEqual(len(kinds), len(self.known_kinds))
            for kind in kinds:
                self.assertIn(kind, self.known_kinds, msg="Kind '{}' not found in: {}".format(kind, self.known_kinds))

        # Check that there is a StructureData that is an input of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.StructureData, tag='structure')
        builder.append(orm.CalculationNode, with_incoming='structure')
        self.assertGreater(len(builder.all()), 0)

        # Check that there is a RemoteData that is the output of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.CalculationNode, tag='parent')
        builder.append(orm.RemoteData, with_incoming='parent')
        self.assertGreater(len(builder.all()), 0)

    @with_temp_dir
    def test_v03_to_newest(self, temp_dir):
        """Test migration of exported files from v0.3 to newest export version"""
        input_file = get_archive_file('export_v0.3.aiida', **self.external_archive)
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Perform the migration
        migrate_archive(input_file, output_file)
        metadata, _ = get_json_files(output_file)
        verify_metadata_version(metadata, version=newest_version)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Do the necessary checks
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(), self.node_count)

        # Verify that CalculationNodes have non-empty attribute dictionaries
        builder = orm.QueryBuilder().append(orm.CalculationNode)
        for [calculation] in builder.iterall():
            self.assertIsInstance(calculation.attributes, dict)
            self.assertNotEqual(len(calculation.attributes), 0)

        # Verify that the StructureData nodes maintained their (same) label, cell, and kinds
        builder = orm.QueryBuilder().append(orm.StructureData)
        self.assertEqual(
            builder.count(),
            self.struct_count,
            msg='There should be {} StructureData, instead {} were/was found'.format(
                self.struct_count, builder.count()
            )
        )
        for structures in builder.all():
            structure = structures[0]
            self.assertEqual(structure.label, self.known_struct_label)
            self.assertEqual(structure.cell, self.known_cell)

        builder = orm.QueryBuilder().append(orm.StructureData, project=['attributes.kinds'])
        for [kinds] in builder.iterall():
            self.assertEqual(len(kinds), len(self.known_kinds))
            for kind in kinds:
                self.assertIn(kind, self.known_kinds, msg="Kind '{}' not found in: {}".format(kind, self.known_kinds))

        # Check that there is a StructureData that is an input of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.StructureData, tag='structure')
        builder.append(orm.CalculationNode, with_incoming='structure')
        self.assertGreater(len(builder.all()), 0)

        # Check that there is a RemoteData that is the output of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.CalculationNode, tag='parent')
        builder.append(orm.RemoteData, with_incoming='parent')
        self.assertGreater(len(builder.all()), 0)

    @with_temp_dir
    def test_v04_to_newest(self, temp_dir):
        """Test migration of exported files from v0.4 to newest export version"""
        input_file = get_archive_file('export_v0.4.aiida', **self.external_archive)
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Perform the migration
        migrate_archive(input_file, output_file)
        metadata, _ = get_json_files(output_file)
        verify_metadata_version(metadata, version=newest_version)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Do the necessary checks
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(), self.node_count + 2)

        # Verify that CalculationNodes have non-empty attribute dictionaries
        builder = orm.QueryBuilder().append(orm.CalculationNode)
        for [calculation] in builder.iterall():
            self.assertIsInstance(calculation.attributes, dict)
            self.assertNotEqual(len(calculation.attributes), 0)

        # Verify that the StructureData nodes maintained their (same) label, cell, and kinds
        builder = orm.QueryBuilder().append(orm.StructureData)
        self.assertEqual(
            builder.count(),
            self.struct_count,
            msg='There should be {} StructureData, instead {} were/was found'.format(
                self.struct_count, builder.count()
            )
        )
        for structures in builder.all():
            structure = structures[0]
            self.assertEqual(structure.label, self.known_struct_label)
            self.assertEqual(structure.cell, self.known_cell)

        builder = orm.QueryBuilder().append(orm.StructureData, project=['attributes.kinds'])
        for [kinds] in builder.iterall():
            self.assertEqual(len(kinds), len(self.known_kinds))
            for kind in kinds:
                self.assertIn(kind, self.known_kinds, msg="Kind '{}' not found in: {}".format(kind, self.known_kinds))

        # Check that there is a StructureData that is an input of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.StructureData, tag='structure')
        builder.append(orm.CalculationNode, with_incoming='structure')
        self.assertGreater(len(builder.all()), 0)

        # Check that there is a RemoteData that is the output of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.CalculationNode, tag='parent')
        builder.append(orm.RemoteData, with_incoming='parent')
        self.assertGreater(len(builder.all()), 0)

    @with_temp_dir
    def test_v05_to_newest(self, temp_dir):
        """Test migration of exported files from v0.5 to newest export version"""
        input_file = get_archive_file('export_v0.5_manual.aiida', **self.external_archive)
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Perform the migration
        migrate_archive(input_file, output_file)
        metadata, _ = get_json_files(output_file)
        verify_metadata_version(metadata, version=newest_version)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Do the necessary checks
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(), self.node_count + 2)

        # Verify that CalculationNodes have non-empty attribute dictionaries
        builder = orm.QueryBuilder().append(orm.CalculationNode)
        for [calculation] in builder.iterall():
            self.assertIsInstance(calculation.attributes, dict)
            self.assertNotEqual(len(calculation.attributes), 0)

        # Verify that the StructureData nodes maintained their (same) label, cell, and kinds
        builder = orm.QueryBuilder().append(orm.StructureData)
        self.assertEqual(
            builder.count(),
            self.struct_count,
            msg='There should be {} StructureData, instead {} were/was found'.format(
                self.struct_count, builder.count()
            )
        )
        for structures in builder.all():
            structure = structures[0]
            self.assertEqual(structure.label, self.known_struct_label)
            self.assertEqual(structure.cell, self.known_cell)

        builder = orm.QueryBuilder().append(orm.StructureData, project=['attributes.kinds'])
        for [kinds] in builder.iterall():
            self.assertEqual(len(kinds), len(self.known_kinds))
            for kind in kinds:
                self.assertIn(kind, self.known_kinds, msg="Kind '{}' not found in: {}".format(kind, self.known_kinds))

        # Check that there is a StructureData that is an input of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.StructureData, tag='structure')
        builder.append(orm.CalculationNode, with_incoming='structure')
        self.assertGreater(len(builder.all()), 0)

        # Check that there is a RemoteData that is the output of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.CalculationNode, tag='parent')
        builder.append(orm.RemoteData, with_incoming='parent')
        self.assertGreater(len(builder.all()), 0)

    @with_temp_dir
    def test_v06_to_newest(self, temp_dir):
        """Test migration of exported files from v0.6 to newest export version"""
        input_file = get_archive_file('export_v0.6_manual.aiida', **self.external_archive)
        output_file = os.path.join(temp_dir, 'output_file.aiida')

        # Perform the migration
        migrate_archive(input_file, output_file)
        metadata, _ = get_json_files(output_file)
        verify_metadata_version(metadata, version=newest_version)

        # Load the migrated file
        import_data(output_file, silent=True)

        # Do the necessary checks
        self.assertEqual(orm.QueryBuilder().append(orm.Node).count(), self.node_count + 2)

        # Verify that CalculationNodes have non-empty attribute dictionaries
        builder = orm.QueryBuilder().append(orm.CalculationNode)
        for [calculation] in builder.iterall():
            self.assertIsInstance(calculation.attributes, dict)
            self.assertNotEqual(len(calculation.attributes), 0)

        # Verify that the StructureData nodes maintained their (same) label, cell, and kinds
        builder = orm.QueryBuilder().append(orm.StructureData)
        self.assertEqual(
            builder.count(),
            self.struct_count,
            msg='There should be {} StructureData, instead {} were/was found'.format(
                self.struct_count, builder.count()
            )
        )
        for structures in builder.all():
            structure = structures[0]
            self.assertEqual(structure.label, self.known_struct_label)
            self.assertEqual(structure.cell, self.known_cell)

        builder = orm.QueryBuilder().append(orm.StructureData, project=['attributes.kinds'])
        for [kinds] in builder.iterall():
            self.assertEqual(len(kinds), len(self.known_kinds))
            for kind in kinds:
                self.assertIn(kind, self.known_kinds, msg="Kind '{}' not found in: {}".format(kind, self.known_kinds))

        # Check that there is a StructureData that is an input of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.StructureData, tag='structure')
        builder.append(orm.CalculationNode, with_incoming='structure')
        self.assertGreater(len(builder.all()), 0)

        # Check that there is a RemoteData that is the output of a CalculationNode
        builder = orm.QueryBuilder()
        builder.append(orm.CalculationNode, tag='parent')
        builder.append(orm.RemoteData, with_incoming='parent')
        self.assertGreater(len(builder.all()), 0)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

from django.db import models, migrations
from aiida.backends.djsite.db.migrations import upgrade_schema_version


REVISION = '1.0.13'
DOWN_REVISION = '1.0.12'


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0012_drop_dblock'),
    ]

    # An amalgamation from django:django/contrib/auth/migrations/
    # these changes are already the default for SQLA at this point
    operations = [
        migrations.AlterField(
            model_name='dbuser',
            name='last_login',
            field=models.DateTimeField(null=True, verbose_name='last login', blank=True),
        ),
        migrations.AlterField(
            model_name='dbuser',
            name='email',
            field=models.EmailField(max_length=254, verbose_name='email address', blank=True),
        ),
        upgrade_schema_version(REVISION, DOWN_REVISION)
    ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Common password and hash generation functions.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import hashlib
try:  # Python3
    from hashlib import blake2b
except ImportError:  # Python2
    from pyblake2 import blake2b
import numbers
import random
import time
import uuid
from datetime import datetime
from operator import itemgetter
from itertools import chain

import six
from six.moves import range
from passlib.context import CryptContext
import pytz

try:  # Python3
    from functools import singledispatch
    from collections import abc, OrderedDict
except ImportError:  # Python2
    from singledispatch import singledispatch
    import collections as abc
    from collections import OrderedDict

from aiida.common.constants import AIIDA_FLOAT_PRECISION

from .folders import Folder

# The prefix of the hashed using pbkdf2_sha256 algorithm in Django
HASHING_PREFIX_DJANGO = 'pbkdf2_sha256'
# The prefix of the hashed using pbkdf2_sha256 algorithm in Passlib
HASHING_PREFIX_PBKDF2_SHA256 = '$pbkdf2-sha256'

# This will never be a valid encoded hash
UNUSABLE_PASSWORD_PREFIX = '!'  # noqa
# Number of random chars to add after UNUSABLE_PASSWORD_PREFIX
UNUSABLE_PASSWORD_SUFFIX_LENGTH = 40

HASHING_KEY = 'HashingKey'

# The key that is used to store the hash in the node extras
_HASH_EXTRA_KEY = '_aiida_hash'

pwd_context = CryptContext(  # pylint: disable=invalid-name
    # The list of hashes that we support
    schemes=['argon2', 'pbkdf2_sha256', 'des_crypt'],
    # The default hashing mechanism
    default='pbkdf2_sha256',

    # We set the number of rounds that should be used...
    pbkdf2_sha256__default_rounds=8000,
)

###################################################################
# THE FOLLOWING WAS TAKEN FROM DJANGO BUT IT CAN BE EASILY REPLACED
###################################################################

# Use the system PRNG if possible
try:
    # pylint: disable=invalid-name
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    import warnings
    warnings.warn('A secure pseudo-random number generator is not available. Falling back to Mersenne Twister.')  # pylint: disable=no-member
    using_sysrandom = False  # pylint: disable=invalid-name


def get_random_string(length=12, allowed_chars='abcdefghijklmnopqrstuvwxyz' 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'):
    """
    Returns a securely generated random string.

    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(hashlib.sha256(('%s%s%s' % (random.getstate(), time.time(), HASHING_KEY)).encode('utf-8')).digest())
    return u''.join(random.choice(allowed_chars) for i in range(length))


BLAKE2B_OPTIONS = {
    'fanout': 0,  # unlimited fanout/depth mode
    'depth': 2,  # has fixed depth of 2
    'digest_size': 32,  # we do not need a cryptographically relevant digest
    'inner_size': 64,  # ... but still use 64 as the inner size
}


def make_hash(object_to_hash, **kwargs):
    """
    Makes a hash from a dictionary, list, tuple or set to any level, that contains
    only other hashable or nonhashable types (including lists, tuples, sets, and
    dictionaries).

    :param object_to_hash: the object to hash

    :returns: a unique hash

    There are a lot of modules providing functionalities to create unique
    hashes for hashable values.
    However, getting hashes for nonhashable items like sets or dictionaries is
    not easily doable because order is not fixed.
    This leads to the peril of getting different hashes for the same
    dictionary.

    This function avoids this by recursing through nonhashable items and
    hashing iteratively. Uses python's sorted function to sort unsorted
    sets and dictionaries by sorting the hashed keys.
    """

    hashes = _make_hash(object_to_hash, **kwargs)  # pylint: disable=assignment-from-no-return

    # use the Unlimited fanout hashing protocol outlined in
    #   https://blake2.net/blake2_20130129.pdf
    final_hash = blake2b(node_depth=1, last_node=True, **BLAKE2B_OPTIONS)

    for sub in hashes:
        final_hash.update(sub)

    # add an empty last leaf node
    final_hash.update(blake2b(node_depth=0, last_node=True, **BLAKE2B_OPTIONS).digest())

    return final_hash.hexdigest()


@singledispatch
def _make_hash(object_to_hash, **_):
    """
    Implementation of the ``make_hash`` function. The hash is created as a
    28 byte integer, and only later converted to a string.
    """
    raise ValueError('Value of type {} cannot be hashed'.format(type(object_to_hash)))


def _single_digest(obj_type, obj_bytes=b''):
    return blake2b(obj_bytes, person=obj_type.encode('ascii'), node_depth=0, **BLAKE2B_OPTIONS).digest()


_END_DIGEST = _single_digest(')')


@_make_hash.register(six.binary_type)
def _(bytes_obj, **kwargs):
    """
    Hash arbitrary binary strings (str in Python 2, bytes in Python 3).
    For compat reason between Python 2 and 3, this gets the same hash-type
    as for unicode in Python 2, resp. str in Python 3."""
    return [_single_digest('str', bytes_obj)]


@_make_hash.register(six.text_type)
def _(val, **kwargs):
    """
    If the type is unicode in Python 2 or a str in Python 3, convert it
    to a str in Python 2 and bytes in Python 3 using the utf-8 encoding.
    """
    return [_single_digest('str', val.encode('utf-8'))]


@_make_hash.register(abc.Sequence)
def _(sequence_obj, **kwargs):
    # unpack the list and use the elements
    return [_single_digest('list(')] + list(chain.from_iterable(_make_hash(i, **kwargs) for i in sequence_obj)
                                           ) + [_END_DIGEST]


@_make_hash.register(abc.Set)
def _(set_obj, **kwargs):
    # turn the set objects into a list of hashes which are always sortable,
    # then return a flattened list of the hashes
    return [_single_digest('set(')] + list(chain.from_iterable(sorted(_make_hash(i, **kwargs) for i in set_obj))
                                          ) + [_END_DIGEST]


@_make_hash.register(abc.Mapping)
def _(mapping, **kwargs):
    """Hashing arbitrary mapping containers (dict, OrderedDict) by first sorting by hashed keys"""

    def hashed_key_mapping():
        for key, value in mapping.items():
            yield (_make_hash(key, **kwargs), value)

    return [_single_digest('dict(')] + list(
        chain.from_iterable(
            (k_digest + _make_hash(val, **kwargs)) for k_digest, val in sorted(hashed_key_mapping(), key=itemgetter(0))
        )
    ) + [_END_DIGEST]


@_make_hash.register(OrderedDict)
def _(mapping, **kwargs):
    """
    Hashing of OrderedDicts

    :param odict_as_unordered: hash OrderedDicts as normal dicts (mostly for testing)
    """

    if kwargs.get('odict_as_unordered', False):
        return _make_hash.registry[abc.Mapping](mapping)

    return ([_single_digest('odict(')] + list(
        chain.from_iterable((_make_hash(key, **kwargs) + _make_hash(val, **kwargs)) for key, val in mapping.items())
    ) + [_END_DIGEST])


@_make_hash.register(numbers.Real)
def _(val, **kwargs):
    """
    Before hashing a float, convert to a string (via rounding) and with a fixed number of digits after the comma.
    Note that the `_singe_digest` requires a bytes object so we need to encode the utf-8 string first
    """
    return [_single_digest('float', float_to_text(val, sig=AIIDA_FLOAT_PRECISION).encode('utf-8'))]


@_make_hash.register(numbers.Complex)
def _(val, **kwargs):
    """
    In case of a complex number, use the same encoding of two floats and join them with a special symbol (a ! here).
    """
    return [
        _single_digest(
            'complex', u'{}!{}'.format(
                float_to_text(val.real, sig=AIIDA_FLOAT_PRECISION), float_to_text(val.imag, sig=AIIDA_FLOAT_PRECISION)
            ).encode('utf-8')
        )
    ]


@_make_hash.register(numbers.Integral)
def _(val, **kwargs):
    """get the hash of the little-endian signed long long representation of the integer"""
    return [_single_digest('int', u'{}'.format(val).encode('utf-8'))]


@_make_hash.register(bool)
def _(val, **kwargs):
    return [_single_digest('bool', b'\x01' if val else b'\x00')]


@_make_hash.register(type(None))
def _(val, **kwargs):
    return [_single_digest('none')]


@_make_hash.register(datetime)
def _(val, **kwargs):
    """hashes the little-endian rep of the float <epoch-seconds>.<subseconds>"""

    # see also https://stackoverflow.com/a/8778548 for an excellent elaboration

    if six.PY2:
        if val.tzinfo is not None and val.utcoffset() is not None:
            val = val.replace(tzinfo=None) - val.utcoffset()
        timestamp = (val - datetime(1970, 1, 1)).total_seconds()
    else:
        if val.tzinfo is None or val.utcoffset() is None:
            val = val.replace(tzinfo=pytz.utc)
        timestamp = val.timestamp()

    return [_single_digest('datetime', float_to_text(timestamp, sig=AIIDA_FLOAT_PRECISION).encode('utf-8'))]


@_make_hash.register(uuid.UUID)
def _(val, **kwargs):
    return [_single_digest('uuid', val.bytes)]


@_make_hash.register(Folder)
def _(folder, **kwargs):
    """
    Hash the content of a Folder object. The name of the folder itself is actually ignored
    :param ignored_folder_content: list of filenames to be ignored for the hashing
    """

    ignored_folder_content = kwargs.get('ignored_folder_content', [])

    def folder_digests(subfolder):
        """traverses the given folder and yields digests for the contained objects"""
        for name, isfile in sorted(subfolder.get_content_list(only_paths=False), key=itemgetter(0)):
            if name in ignored_folder_content:
                continue

            if isfile:
                yield _single_digest('fname', name.encode('utf-8'))
                with subfolder.open(name, mode='rb') as fhandle:
                    yield _single_digest('fcontent', fhandle.read())
            else:
                yield _single_digest('dir(', name.encode('utf-8'))
                for digest in folder_digests(subfolder.get_subfolder(name)):
                    yield digest
                yield _END_DIGEST

    return [_single_digest('folder')] + [d for d in folder_digests(folder)]


def float_to_text(value, sig):
    """
    Convert float to text string for computing hash.
    Preseve up to N significant number given by sig.

    :param value: the float value to convert
    :param sig: choose how many digits after the comma should be output
    """
    fmt = u'{{:.{}g}}'.format(sig)
    return fmt.format(value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Translator for code
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.restapi.translator.nodes.data import DataTranslator


class CodeTranslator(DataTranslator):
    """
    Translator relative to resource 'codes' and aiida class Code
    """

    # A label associated to the present class (coincides with the resource name)
    __label__ = 'codes'
    # The AiiDA class one-to-one associated to the present class
    from aiida.orm import Code
    _aiida_class = Code
    # The string name of the AiiDA class
    _aiida_type = 'data.code.Code'

    _result_type = __label__

    def __init__(self, **kwargs):
        """
        Initialise the parameters.
        Create the basic query_help
        """
        super(CodeTranslator, self).__init__(Class=self.__class__, **kwargs)

    @staticmethod
    def get_visualization_data(node, visformat=None):
        """
        Generic function extented for codes data. Currently
        it is not implemented.

        :param node: node object that has to be visualized
        :param visformat: visualization format
        :returns: raise RestFeatureNotAvailable exception
        """

        from aiida.restapi.common.exceptions import RestFeatureNotAvailable

        raise RestFeatureNotAvailable('This endpoint is not available for Codes.')

    @staticmethod
    def get_downloadable_data(node, download_format=None):
        """
        Generic function extented for codes data. Currently
        it is not implemented.

        :param node: node object that has to be downloaded
        :param download_format: file extension format
        :returns: raise RestFeatureNotAvailable exception
        """

        from aiida.restapi.common.exceptions import RestFeatureNotAvailable

        raise RestFeatureNotAvailable('This endpoint is not available for Codes.')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Django query builder"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import uuid
import six

from aldjemy import core
# Remove when https://github.com/PyCQA/pylint/issues/1931 is fixed
# pylint: disable=no-name-in-module, import-error
from sqlalchemy_utils.types.choice import Choice
from sqlalchemy import and_, or_, not_, case
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import FunctionElement
from sqlalchemy.types import Integer, Float, Boolean, DateTime

import aiida.backends.djsite.db.models as djmodels
from aiida.common.exceptions import InputValidationError
from aiida.orm.implementation.querybuilder import BackendQueryBuilder


class jsonb_array_length(FunctionElement):  # pylint: disable=invalid-name
    # pylint: disable=too-few-public-methods
    name = 'jsonb_array_len'


@compiles(jsonb_array_length)
def compile(element, compiler, **_kw):  # pylint: disable=function-redefined, redefined-builtin
    """
    Get length of array defined in a JSONB column
    """
    return 'jsonb_array_length(%s)' % compiler.process(element.clauses)


class array_length(FunctionElement):  # pylint: disable=invalid-name
    # pylint: disable=too-few-public-methods
    name = 'array_len'


@compiles(array_length)
def compile(element, compiler, **_kw):  # pylint: disable=function-redefined
    """
    Get length of array defined in a JSONB column
    """
    return 'array_length(%s)' % compiler.process(element.clauses)


class jsonb_typeof(FunctionElement):  # pylint: disable=invalid-name
    # pylint: disable=too-few-public-methods
    name = 'jsonb_typeof'


@compiles(jsonb_typeof)
def compile(element, compiler, **_kw):  # pylint: disable=function-redefined
    """
    Get length of array defined in a JSONB column
    """
    return 'jsonb_typeof(%s)' % compiler.process(element.clauses)


class DjangoQueryBuilder(BackendQueryBuilder):
    """Django query builder"""

    # pylint: disable=too-many-public-methods,no-member

    def __init__(self, backend):
        BackendQueryBuilder.__init__(self, backend)

    @property
    def Node(self):
        return djmodels.DbNode.sa

    @property
    def Link(self):
        return djmodels.DbLink.sa

    @property
    def Computer(self):
        return djmodels.DbComputer.sa

    @property
    def User(self):
        return djmodels.DbUser.sa

    @property
    def Group(self):
        return djmodels.DbGroup.sa

    @property
    def AuthInfo(self):
        return djmodels.DbAuthInfo.sa

    @property
    def Comment(self):
        return djmodels.DbComment.sa

    @property
    def Log(self):
        return djmodels.DbLog.sa

    @property
    def table_groups_nodes(self):
        return core.Cache.meta.tables['db_dbgroup_dbnodes']

    def get_filter_expr(self, operator, value, attr_key, is_attribute, alias=None, column=None, column_name=None):
        """
        Applies a filter on the alias given.
        Expects the alias of the ORM-class on which to filter, and filter_spec.
        Filter_spec contains the specification on the filter.
        Expects:

        :param operator: The operator to apply, see below for further details
        :param value:
            The value for the right side of the expression,
            the value you want to compare with.

        :param path: The path leading to the value

        :param attr_key: Boolean, whether the value is in a json-column,
            or in an attribute like table.


        Implemented and valid operators:

        *   for any type:
            *   ==  (compare single value, eg: '==':5.0)
            *   in    (compare whether in list, eg: 'in':[5, 6, 34]
        *  for floats and integers:
            *   >
            *   <
            *   <=
            *   >=
        *  for strings:
            *   like  (case - sensitive), for example
                'like':'node.calc.%'  will match node.calc.relax and
                node.calc.RELAX and node.calc. but
                not node.CALC.relax
            *   ilike (case - unsensitive)
                will also match node.CaLc.relax in the above example

            .. note::
                The character % is a reserved special character in SQL,
                and acts as a wildcard. If you specifically
                want to capture a ``%`` in the string, use: ``_%``

        *   for arrays and dictionaries (only for the
            SQLAlchemy implementation):

            *   contains: pass a list with all the items that
                the array should contain, or that should be among
                the keys, eg: 'contains': ['N', 'H'])
            *   has_key: pass an element that the list has to contain
                or that has to be a key, eg: 'has_key':'N')

        *  for arrays only (SQLAlchemy version):
            *   of_length
            *   longer
            *   shorter

        All the above filters invoke a negation of the
        expression if preceded by **~**::

            # first example:
            filter_spec = {
                'name' : {
                    '~in':[
                        'halle',
                        'lujah'
                    ]
                } # Name not 'halle' or 'lujah'
            }

            # second example:
            filter_spec =  {
                'id' : {
                    '~==': 2
                }
            } # id is not 2
        """
        # pylint: disable=too-many-branches,too-many-arguments
        # pylint: disable=too-many-branches,too-many-arguments

        expr = None
        if operator.startswith('~'):
            negation = True
            operator = operator.lstrip('~')
        elif operator.startswith('!'):
            negation = True
            operator = operator.lstrip('!')
        else:
            negation = False
        if operator in ('longer', 'shorter', 'of_length'):
            if not isinstance(value, int):
                raise InputValidationError('You have to give an integer when comparing to a length')
        elif operator in ('like', 'ilike'):
            if not isinstance(value, six.string_types):
                raise InputValidationError(
                    'Value for operator {} has to be a string (you gave {})'
                    ''.format(operator, value)
                )
        elif operator == 'in':
            value_type_set = set(type(i) for i in value)
            if len(value_type_set) > 1:
                raise InputValidationError('{}  contains more than one type'.format(value))
            elif not value_type_set:
                raise InputValidationError('{}  contains is an empty list'.format(value))
        elif operator in ('and', 'or'):
            expressions_for_this_path = []
            for filter_operation_dict in value:
                for newoperator, newvalue in filter_operation_dict.items():
                    expressions_for_this_path.append(
                        self.get_filter_expr(
                            newoperator,
                            newvalue,
                            attr_key=attr_key,
                            is_attribute=is_attribute,
                            alias=alias,
                            column=column,
                            column_name=column_name
                        )
                    )
            if operator == 'and':
                expr = and_(*expressions_for_this_path)
            elif operator == 'or':
                expr = or_(*expressions_for_this_path)

        if expr is None:
            if is_attribute:
                expr = self.get_filter_expr_from_attributes(
                    operator, value, attr_key, column=column, column_name=column_name, alias=alias
                )
            else:
                if column is None:
                    if (alias is None) and (column_name is None):
                        raise Exception('I need to get the column but do not know the alias and the column name')
                    column = self.get_column(column_name, alias)
                expr = self.get_filter_expr_from_column(operator, value, column)
        if negation:
            return not_(expr)
        return expr

    @staticmethod
    def get_session():
        from aiida.manage.configuration import get_config
        from aiida.backends.sqlalchemy import reset_session
        from aiida.backends.sqlalchemy import get_scoped_session

        if get_scoped_session() is None:
            config = get_config()
            profile = config.current_profile
            reset_session(profile)

        return get_scoped_session()

    def modify_expansions(self, alias, expansions):
        """
        For django, there are no additional expansions for now, so
        I am returning an empty list
        """
        return expansions

    def get_filter_expr_from_attributes(self, operator, value, attr_key, column=None, column_name=None, alias=None):
        # Too many everything!
        # pylint: disable=too-many-branches, too-many-arguments, too-many-statements

        def cast_according_to_type(path_in_json, value):
            """Cast the value according to the type"""
            if isinstance(value, bool):
                type_filter = jsonb_typeof(path_in_json) == 'boolean'
                casted_entity = path_in_json.astext.cast(Boolean)
            elif isinstance(value, (int, float)):
                type_filter = jsonb_typeof(path_in_json) == 'number'
                casted_entity = path_in_json.astext.cast(Float)
            elif isinstance(value, dict) or value is None:
                type_filter = jsonb_typeof(path_in_json) == 'object'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            elif isinstance(value, dict):
                type_filter = jsonb_typeof(path_in_json) == 'array'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            elif isinstance(value, six.string_types):
                type_filter = jsonb_typeof(path_in_json) == 'string'
                casted_entity = path_in_json.astext
            elif value is None:
                type_filter = jsonb_typeof(path_in_json) == 'null'
                casted_entity = path_in_json.astext.cast(JSONB)  # BOOLEANS?
            else:
                raise TypeError('Unknown type {}'.format(type(value)))
            return type_filter, casted_entity

        if column is None:
            column = self.get_column(column_name, alias)

        database_entity = column[tuple(attr_key)]
        if operator == '==':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity == value)], else_=False)
        elif operator == '>':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity > value)], else_=False)
        elif operator == '<':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity < value)], else_=False)
        elif operator in ('>=', '=>'):
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity >= value)], else_=False)
        elif operator in ('<=', '=<'):
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity <= value)], else_=False)
        elif operator == 'of_type':
            # http://www.postgresql.org/docs/9.5/static/functions-json.html
            #  Possible types are object, array, string, number, boolean, and null.
            valid_types = ('object', 'array', 'string', 'number', 'boolean', 'null')
            if value not in valid_types:
                raise InputValidationError(
                    'value {} for of_type is not among valid types\n'
                    '{}'.format(value, valid_types)
                )
            expr = jsonb_typeof(database_entity) == value
        elif operator == 'like':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity.like(value))], else_=False)
        elif operator == 'ilike':
            type_filter, casted_entity = cast_according_to_type(database_entity, value)
            expr = case([(type_filter, casted_entity.ilike(value))], else_=False)
        elif operator == 'in':
            type_filter, casted_entity = cast_according_to_type(database_entity, value[0])
            expr = case([(type_filter, casted_entity.in_(value))], else_=False)
        elif operator == 'contains':
            expr = database_entity.cast(JSONB).contains(value)
        elif operator == 'has_key':
            expr = database_entity.cast(JSONB).has_key(value)  # noqa
        elif operator == 'of_length':
            expr = case([
                (jsonb_typeof(database_entity) == 'array', jsonb_array_length(database_entity.cast(JSONB)) == value)
            ],
                        else_=False)

        elif operator == 'longer':
            expr = case([
                (jsonb_typeof(database_entity) == 'array', jsonb_array_length(database_entity.cast(JSONB)) > value)
            ],
                        else_=False)
        elif operator == 'shorter':
            expr = case([
                (jsonb_typeof(database_entity) == 'array', jsonb_array_length(database_entity.cast(JSONB)) < value)
            ],
                        else_=False)
        else:
            raise InputValidationError('Unknown operator {} for filters in JSON field'.format(operator))
        return expr

    def get_projectable_attribute(self, alias, column_name, attrpath, cast=None, **kwargs):  # pylint: disable=redefined-outer-name
        """
        :returns: An attribute store in a JSON field of the give column
        """
        entity = self.get_column(column_name, alias)[attrpath]
        if cast is None:
            entity = entity
        elif cast == 'f':
            entity = entity.astext.cast(Float)
        elif cast == 'i':
            entity = entity.astext.cast(Integer)
        elif cast == 'b':
            entity = entity.astext.cast(Boolean)
        elif cast == 't':
            entity = entity.astext
        elif cast == 'j':
            entity = entity.astext.cast(JSONB)
        elif cast == 'd':
            entity = entity.astext.cast(DateTime)
        else:
            raise InputValidationError('Unkown casting key {}'.format(cast))
        return entity

    def get_aiida_res(self, key, res):
        """
        Some instance returned by ORM (django or SA) need to be converted
        to Aiida instances (eg nodes). Choice (sqlalchemy_utils)
        will return their value

        :param key: The key
        :param res: the result returned by the query

        :returns: an aiida-compatible instance
        """
        if isinstance(res, Choice):
            result = res.value
        elif isinstance(res, uuid.UUID):
            result = six.text_type(res)
        else:
            try:
                result = self._backend.get_backend_entity(res)
            except TypeError:
                result = res

        return result

    def yield_per(self, query, batch_size):
        """
        :param count: Number of rows to yield per step

        Yields *count* rows at a time

        :returns: a generator
        """
        from django.db import transaction
        with transaction.atomic():
            return query.yield_per(batch_size)

    def count(self, query):

        from django.db import transaction
        with transaction.atomic():
            return query.count()

    def first(self, query):
        """
        Executes query in the backend asking for one instance.

        :returns: One row of aiida results
        """
        from django.db import transaction
        with transaction.atomic():
            return query.first()

    def iterall(self, query, batch_size, tag_to_index_dict):
        from django.db import transaction
        if not tag_to_index_dict:
            raise ValueError('Got an empty dictionary: {}'.format(tag_to_index_dict))

        with transaction.atomic():
            results = query.yield_per(batch_size)

            if len(tag_to_index_dict) == 1:
                # Sqlalchemy, for some strange reason, does not return a list of lists
                # if you have provided an ormclass

                if list(tag_to_index_dict.values()) == ['*']:
                    for rowitem in results:
                        yield [self.get_aiida_res(tag_to_index_dict[0], rowitem)]
                else:
                    for rowitem, in results:
                        yield [self.get_aiida_res(tag_to_index_dict[0], rowitem)]
            elif len(tag_to_index_dict) > 1:
                for resultrow in results:
                    yield [
                        self.get_aiida_res(tag_to_index_dict[colindex], rowitem)
                        for colindex, rowitem in enumerate(resultrow)
                    ]

    def iterdict(self, query, batch_size, tag_to_projected_properties_dict, tag_to_alias_map):
        from django.db import transaction

        def get_table_name(aliased_class):
            """ Returns the table name given an Aliased class based on Aldjemy"""
            return aliased_class._aliased_insp._target.table.name  # pylint: disable=protected-access

        nr_items = sum(len(v) for v in tag_to_projected_properties_dict.values())

        if not nr_items:
            raise ValueError('Got an empty dictionary')

        # Wrapping everything in an atomic transaction:
        with transaction.atomic():
            results = query.yield_per(batch_size)
            # Two cases: If one column was asked, the database returns a matrix of rows * columns:
            if nr_items > 1:
                for this_result in results:
                    yield {
                        tag: {
                            self.get_corresponding_property(
                                get_table_name(tag_to_alias_map[tag]), attrkey, self.inner_to_outer_schema
                            ): self.get_aiida_res(attrkey, this_result[index_in_sql_result])
                            for attrkey, index_in_sql_result in projected_entities_dict.items()
                        } for tag, projected_entities_dict in tag_to_projected_properties_dict.items()
                    }
            elif nr_items == 1:
                # I this case, sql returns a  list, where each listitem is the result
                # for one row. Here I am converting it to a list of lists (of length 1)
                if [v for entityd in tag_to_projected_properties_dict.values() for v in entityd.keys()] == ['*']:
                    for this_result in results:
                        yield {
                            tag: {
                                self.get_corresponding_property(
                                    get_table_name(tag_to_alias_map[tag]), attrkey, self.inner_to_outer_schema
                                ): self.get_aiida_res(attrkey, this_result)
                                for attrkey, position in projected_entities_dict.items()
                            } for tag, projected_entities_dict in tag_to_projected_properties_dict.items()
                        }
                else:
                    for this_result, in results:
                        yield {
                            tag: {
                                self.get_corresponding_property(
                                    get_table_name(tag_to_alias_map[tag]), attrkey, self.inner_to_outer_schema
                                ): self.get_aiida_res(attrkey, this_result)
                                for attrkey, position in projected_entities_dict.items()
                            } for tag, projected_entities_dict in tag_to_projected_properties_dict.items()
                        }

    def get_column_names(self, alias):
        """
        Given the backend specific alias, return the column names that correspond to the aliased table.
        """
        # pylint: disable=protected-access
        return [
            str(c).replace(alias._aliased_insp.class_.table.name + '.', '')
            for c in alias._aliased_insp.class_.table._columns._all_columns
        ]



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from alembic import command
from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.script import ScriptDirectory

from sqlalchemy.orm.exc import NoResultFound

from aiida.backends import sqlalchemy as sa
from aiida.backends.sqlalchemy import get_scoped_session
from aiida.backends.utils import validate_attribute_key, SettingsManager, Setting
from aiida.common import NotExistent


ALEMBIC_FILENAME = 'alembic.ini'
ALEMBIC_REL_PATH = 'migrations'


class SqlaSettingsManager(SettingsManager):
    """Class to get, set and delete settings from the `DbSettings` table."""

    table_name = 'db_dbsetting'

    def validate_table_existence(self):
        """Verify that the `DbSetting` table actually exists.

        :raises: `~aiida.common.exceptions.NotExistent` if the settings table does not exist
        """
        from sqlalchemy.engine import reflection
        inspector = reflection.Inspector.from_engine(get_scoped_session().bind)
        if self.table_name not in inspector.get_table_names():
            raise NotExistent('the settings table does not exist')

    def get(self, key):
        """Return the setting with the given key.

        :param key: the key identifying the setting
        :return: Setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """
        from aiida.backends.sqlalchemy.models.settings import DbSetting
        self.validate_table_existence()

        try:
            setting = get_scoped_session().query(DbSetting).filter_by(key=key).one()
        except NoResultFound:
            raise NotExistent('setting `{}` does not exist'.format(key))

        return Setting(key, setting.getvalue(), setting.description, setting.time)

    def set(self, key, value, description=None):
        """Return the settings with the given key.

        :param key: the key identifying the setting
        :param value: the value for the setting
        :param description: optional setting description
        """
        from aiida.backends.sqlalchemy.models.settings import DbSetting
        self.validate_table_existence()
        validate_attribute_key(key)

        other_attribs = dict()
        if description is not None:
            other_attribs['description'] = description

        DbSetting.set_value(key, value, other_attribs=other_attribs)

    def delete(self, key):
        """Delete the setting with the given key.

        :param key: the key identifying the setting
        :raises: `~aiida.common.exceptions.NotExistent` if the settings does not exist
        """
        from aiida.backends.sqlalchemy.models.settings import DbSetting
        self.validate_table_existence()

        try:
            setting = get_scoped_session().query(DbSetting).filter_by(key=key).one()
            setting.delete()
        except NoResultFound:
            raise NotExistent('setting `{}` does not exist'.format(key))



def flag_modified(instance, key):
    """Wrapper around `sqlalchemy.orm.attributes.flag_modified` to correctly dereference utils.ModelWrapper

    Since SqlAlchemy 1.2.12 (and maybe earlier but not in 1.0.19) the flag_modified function will check that the
    key is actually present in the instance or it will except. If we pass a model instance, wrapped in the ModelWrapper
    the call will raise an InvalidRequestError. In this function that wraps the flag_modified of SqlAlchemy, we
    derefence the model instance if the passed instance is actually wrapped in the ModelWrapper.
    """
    from sqlalchemy.orm.attributes import flag_modified as flag_modified_sqla
    from aiida.orm.implementation.sqlalchemy.utils import ModelWrapper

    if isinstance(instance, ModelWrapper):
        instance = instance._model

    flag_modified_sqla(instance, key)


def load_dbenv(profile):
    """Load the database environment and ensure that the code and database schema versions are compatible.

    :param profile: the string with the profile to use
    """
    _load_dbenv_noschemacheck(profile)
    check_schema_version(profile_name=profile.name)


def _load_dbenv_noschemacheck(profile):
    """Load the database environment without checking that code and database schema versions are compatible.

    This should ONLY be used internally, inside load_dbenv, and for schema migrations. DO NOT USE OTHERWISE!

    :param profile: instance of `Profile` whose database to load
    """
    sa.reset_session(profile)


def unload_dbenv():
    """Unload the database environment, which boils down to destroying the current engine and session."""
    if sa.ENGINE is not None:
        sa.ENGINE.dispose()
    sa.SCOPED_SESSION_CLASS = None


_aiida_autouser_cache = None


# XXX the code here isn't different from the one use in Django. We may be able
# to refactor it in some way
def install_tc(session):
    """
    Install the transitive closure table with SqlAlchemy.
    """
    links_table_name = 'db_dblink'
    links_table_input_field = 'input_id'
    links_table_output_field = 'output_id'
    closure_table_name = 'db_dbpath'
    closure_table_parent_field = 'parent_id'
    closure_table_child_field = 'child_id'

    session.execute(get_pg_tc(links_table_name, links_table_input_field,
                              links_table_output_field, closure_table_name,
                              closure_table_parent_field,
                              closure_table_child_field))


def get_pg_tc(links_table_name,
              links_table_input_field,
              links_table_output_field,
              closure_table_name,
              closure_table_parent_field,
              closure_table_child_field):
    """
    Return the transitive closure table template
    """
    from string import Template

    pg_tc = Template("""

DROP TRIGGER IF EXISTS autoupdate_tc ON $links_table_name;
DROP FUNCTION IF EXISTS update_tc();

CREATE OR REPLACE FUNCTION update_tc()
  RETURNS trigger AS
$$BODY$$
DECLARE

    new_id INTEGER;
    old_id INTEGER;
    num_rows INTEGER;

BEGIN

  IF tg_op = 'INSERT' THEN

    IF EXISTS (
      SELECT Id FROM $closure_table_name
      WHERE $closure_table_parent_field = new.$links_table_input_field
         AND $closure_table_child_field = new.$links_table_output_field
         AND depth = 0
         )
    THEN
      RETURN null;
    END IF;

    IF new.$links_table_input_field = new.$links_table_output_field
    OR EXISTS (
      SELECT id FROM $closure_table_name
        WHERE $closure_table_parent_field = new.$links_table_output_field
        AND $closure_table_child_field = new.$links_table_input_field
        )
    THEN
      RETURN null;
    END IF;

    INSERT INTO $closure_table_name (
         $closure_table_parent_field,
         $closure_table_child_field,
         depth)
      VALUES (
         new.$links_table_input_field,
         new.$links_table_output_field,
         0);

    new_id := lastval();

    UPDATE $closure_table_name
      SET entry_edge_id = new_id
        , exit_edge_id = new_id
        , direct_edge_id = new_id
      WHERE id = new_id;


    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT id
         , new_id
         , new_id
         , $closure_table_parent_field
         , new.$links_table_output_field
         , depth + 1
        FROM $closure_table_name
        WHERE $closure_table_child_field = new.$links_table_input_field;


    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT new_id
        , new_id
        , id
        , new.$links_table_input_field
        , $closure_table_child_field
        , depth + 1
        FROM $closure_table_name
        WHERE $closure_table_parent_field = new.$links_table_output_field;

    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT A.id
        , new_id
        , B.id
        , A.$closure_table_parent_field
        , B.$closure_table_child_field
        , A.depth + B.depth + 2
     FROM $closure_table_name A
        CROSS JOIN $closure_table_name B
     WHERE A.$closure_table_child_field = new.$links_table_input_field
       AND B.$closure_table_parent_field = new.$links_table_output_field;

  END IF;

  IF tg_op = 'DELETE' THEN

    IF NOT EXISTS(
        SELECT id FROM $closure_table_name
        WHERE $closure_table_parent_field = old.$links_table_input_field
        AND $closure_table_child_field = old.$links_table_output_field AND
        depth = 0 )
    THEN
        RETURN NULL;
    END IF;

    CREATE TABLE PurgeList (Id int);

    INSERT INTO PurgeList
      SELECT id FROM $closure_table_name
          WHERE $closure_table_parent_field = old.$links_table_input_field
        AND $closure_table_child_field = old.$links_table_output_field AND
        depth = 0;

    WHILE (1 = 1)
    loop

      INSERT INTO PurgeList
        SELECT id FROM $closure_table_name
          WHERE depth > 0
          AND ( entry_edge_id IN ( SELECT Id FROM PurgeList )
          OR direct_edge_id IN ( SELECT Id FROM PurgeList )
          OR exit_edge_id IN ( SELECT Id FROM PurgeList ) )
          AND Id NOT IN (SELECT Id FROM PurgeList );

      GET DIAGNOSTICS num_rows = ROW_COUNT;
      if (num_rows = 0) THEN
        EXIT;
      END IF;
    end loop;

    DELETE FROM $closure_table_name WHERE Id IN ( SELECT Id FROM PurgeList);
    DROP TABLE PurgeList;

  END IF;

  RETURN NULL;

END
$$BODY$$
  LANGUAGE plpgsql VOLATILE
  COST 100;


CREATE TRIGGER autoupdate_tc
  AFTER INSERT OR DELETE OR UPDATE
  ON $links_table_name FOR each ROW
  EXECUTE PROCEDURE update_tc();

""")
    return pg_tc.substitute(links_table_name=links_table_name,
                            links_table_input_field=links_table_input_field,
                            links_table_output_field=links_table_output_field,
                            closure_table_name=closure_table_name,
                            closure_table_parent_field=closure_table_parent_field,
                            closure_table_child_field=closure_table_child_field)


def migrate_database(alembic_cfg=None):
    """Migrate the database to the latest schema version.

    :param config: alembic configuration to use, will use default if not provided
    """
    from aiida.backends import sqlalchemy as sa

    if alembic_cfg is None:
        alembic_cfg = get_alembic_conf()

    with sa.ENGINE.connect() as connection:
        alembic_cfg.attributes['connection'] = connection
        command.upgrade(alembic_cfg, 'head')


def check_schema_version(profile_name):
    """
    Check if the version stored in the database is the same of the version of the code.

    :raise aiida.common.ConfigurationError: if the two schema versions do not match
    """
    from aiida.backends import sqlalchemy as sa
    from aiida.common.exceptions import ConfigurationError

    alembic_cfg = get_alembic_conf()

    # Getting the version of the code and the database, reusing the existing engine (initialized by AiiDA)
    with sa.ENGINE.begin() as connection:
        alembic_cfg.attributes['connection'] = connection
        code_schema_version = get_migration_head(alembic_cfg)
        db_schema_version = get_db_schema_version(alembic_cfg)

    if code_schema_version != db_schema_version:
        raise ConfigurationError('Database schema version {} is outdated compared to the code schema version {}\n'
                                 'Before you upgrade, make sure all calculations and workflows have finished running.\n'
                                 'If this is not the case, revert the code to the previous version and finish them first.\n'
                                 'To migrate the database to the current version, run the following commands:'
                                 '\n  verdi -p {} daemon stop\n  verdi -p {} database migrate'.format(
                                    db_schema_version, code_schema_version, profile_name, profile_name))



def get_migration_head(config):
    """
    This function returns the head of the migration scripts.
    :param config: The alembic configuration.
    :return: The version of the head.
    """
    script = ScriptDirectory.from_config(config)
    return script.get_current_head()


def get_db_schema_version(config):
    """
    This function returns the current version of the database.
    :param config: The alembic configuration.
    :return: The version of the database.
    """
    if config is None:
        return None

    script = ScriptDirectory.from_config(config)

    def get_db_version(rev, _):
        if isinstance(rev, tuple) and len(rev) > 0:
            config.attributes['rev'] = rev[0]
        else:
            config.attributes['rev'] = None

        return []

    with EnvironmentContext(
            config,
            script,
            fn=get_db_version
    ):
        script.run_env()
        return config.attributes['rev']


def get_alembic_conf():
    """
    This function returns the alembic configuration file contents by doing
    the necessary updates in the 'script_location' name.
    :return: The alembic configuration.
    """
    # Constructing the alembic full path & getting the configuration
    import os
    dir_path = os.path.dirname(os.path.realpath(__file__))
    alembic_fpath = os.path.join(dir_path, ALEMBIC_FILENAME)
    alembic_cfg = Config(alembic_fpath)

    # Set the alembic script directory location
    alembic_dpath = os.path.join(dir_path, ALEMBIC_REL_PATH)
    alembic_cfg.set_main_option('script_location', alembic_dpath)

    return alembic_cfg


def delete_nodes_and_connections_sqla(pks_to_delete):
    """
    Delete all nodes corresponding to pks in the input.
    :param pks_to_delete: A list, tuple or set of pks that should be deleted.
    """
    from aiida.backends import sqlalchemy as sa
    from aiida.backends.sqlalchemy.models.node import DbNode, DbLink
    from aiida.backends.sqlalchemy.models.group import table_groups_nodes

    session = sa.get_scoped_session()
    try:
        # I am first making a statement to delete the membership of these nodes to groups.
        # Since table_groups_nodes is a sqlalchemy.schema.Table, I am using expression language to compile
        # a stmt to be executed by the session. It works, but it's not nice that two different ways are used!
        # Can this be changed?
        stmt = table_groups_nodes.delete().where(table_groups_nodes.c.dbnode_id.in_(list(pks_to_delete)))
        session.execute(stmt)
        # First delete links, then the Nodes, since we are not cascading deletions.
        # Here I delete the links coming out of the nodes marked for deletion.
        session.query(DbLink).filter(DbLink.input_id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Here I delete the links pointing to the nodes marked for deletion.
        session.query(DbLink).filter(DbLink.output_id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Now I am deleting the nodes
        session.query(DbNode).filter(DbNode.id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Here I commit this scoped session!
        session.commit()
    except Exception as e:
        # If there was any exception, I roll back the session.
        session.rollback()
        raise e
    finally:
        session.close()
